#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/*
 * parse interface address messages
 */ 
int parse_ifamsg(struct nlmsghdr *nlh)
{
	struct ifaddrmsg *ifam = (struct ifaddrmsg *)NLMSG_DATA(nlh);
	int ifam_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);
	struct rtattr *ifa[__IFA_MAX];
	char paddr[INET6_ADDRSTRLEN] = "";
	char laddr[INET6_ADDRSTRLEN] = "";
	char ifname[IFNAMSIZ] = "";
	char iflabel[IFNAMSIZ] = "";
	char msg[MAX_MSG_SIZE] = "";
	char *mp = msg;
	int nmask = -1;
	unsigned short iftype = -1;

	/* parse interface address attributes */
	parse_ifaddr(ifa, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ifamsg(ifam, ifa, ifam_len);
	}

	/* check address family & set message type */
	if(ifam->ifa_family == AF_INET)
		mp += sprintf(mp, "[%-12s] ", "IPv4 ADDRESS");
	else if(ifam->ifa_family == AF_INET6)
		mp += sprintf(mp, "[%-12s] ", "IPv6 ADDRESS");
	else {
		rec_log("[%-12s] %s: unknown address family: %d",
			"ERROR", __func__, ifam->ifa_family);
		return(1);
	}

	/* get interface index & name */
	if(nlh->nlmsg_type == RTM_NEWADDR) {
		if_indextoname(ifam->ifa_index, ifname);
		mp += sprintf(mp, "address added: ");
	} else if(nlh->nlmsg_type == RTM_DELADDR) {
		if(!if_indextoname_from_iflist(ifam->ifa_index, ifname))
			if_indextoname_from_ifhist(ifam->ifa_index, ifname);
		mp += sprintf(mp, "address deleted: ");
	} else {
		rec_log("[%-12s] %s: unknown nlmsg_type: %d",
			"ERROR", __func__, nlh->nlmsg_type);
		return(1);
	}

	/* get interface type */
	iftype = get_type_from_iflist(ifam->ifa_index);
	if(!iftype)
		iftype = get_type_from_ifhist(ifam->ifa_index);

	/* get prefix address */
	if(ifa[IFA_ADDRESS])
		inet_ntop(ifam->ifa_family, RTA_DATA(ifa[IFA_ADDRESS]), paddr, sizeof(paddr));

	/* get local interface address */
	if(ifa[IFA_LOCAL])
		inet_ntop(ifam->ifa_family, RTA_DATA(ifa[IFA_LOCAL]), laddr, sizeof(laddr));
	
	/* whether interface type is P2P */
	if(ifa[IFA_LOCAL] && strcmp(laddr, paddr))
		mp += sprintf(mp, "if=%s local=%s/%d peer=%s/%d",
			ifname, laddr, ifam->ifa_prefixlen, paddr, ifam->ifa_prefixlen);
	else
		mp += sprintf(mp, "if=%s ipaddr=%s/%d",
			ifname, paddr, ifam->ifa_prefixlen);

	/* get scope of address */
	switch(ifam->ifa_scope) {
		case RT_SCOPE_UNIVERSE:
			mp += sprintf(mp, " scope=global");
			break;
		case RT_SCOPE_SITE:
			mp += sprintf(mp, " scope=site");
			break;
		case RT_SCOPE_LINK:
			mp += sprintf(mp, " scope=link");
			break;
		case RT_SCOPE_HOST:
			mp += sprintf(mp, " scope=host");
			break;
		case RT_SCOPE_NOWHERE:
			mp += sprintf(mp, " scope=none");
			break;
		default:
			mp += sprintf(mp, " scope=%d", ifam->ifa_scope);
	}

	/* get interface label */
	if(ifa[IFA_LABEL]) {
		memcpy(iflabel, RTA_DATA(ifa[IFA_LABEL]), IFNAMSIZ);
		if(strcmp(iflabel, ifname))
			mp += sprintf(mp, " label=%s", iflabel);
	}

	/* logging interface address message */
	rec_log("%s", msg);

	return(0);
}

/*
 * debug interface address messages
 */ 
void debug_ifamsg(struct ifaddrmsg *ifam, struct rtattr *ifa[], int ifam_len)
{
	/* debug ifaddrmsg */
	char ifa_flags_list[MAX_STR_SIZE] = "";
	char ifa_scope_name[MAX_STR_SIZE] = "";
	char ifname[IFNAMSIZ] = "";

	rec_dbg("*********************************************************************");
	rec_dbg("[ ifaddrmsg(%d) ]",
		NLMSG_ALIGN(sizeof(struct ifaddrmsg)));
	rec_dbg("    ifa_family(%d): %d(%s)",
		sizeof(ifam->ifa_family), ifam->ifa_family, debug_n2s_af_type(ifam->ifa_family));
	rec_dbg("    ifa_prefixlen(%d): %d",
		sizeof(ifam->ifa_prefixlen), ifam->ifa_prefixlen);
	debug_n2s_ifa_flags(ifam->ifa_flags, ifa_flags_list);
	rec_dbg("    ifa_flags(%d): 0x%x(%s)",
		sizeof(ifam->ifa_flags), ifam->ifa_flags, ifa_flags_list);
	rec_dbg("    ifa_scope(%d): %d(%s)",
		sizeof(ifam->ifa_scope), ifam->ifa_scope, debug_n2s_ifa_scope(ifam->ifa_scope));
	if_indextoname(ifam->ifa_index, ifname);
	rec_dbg("    ifa_index(%d): %u(%s)",
		sizeof(ifam->ifa_index), ifam->ifa_index, ifname);

	/* debug interface address attributes */
	rec_dbg("*********************************************************************");
	rec_dbg("[ ifaddrmsg attributes(%d) ]",
		NLMSG_ALIGN(ifam_len - NLMSG_ALIGN(sizeof(struct ifaddrmsg))));

	/* debug prefix address */
	if(ifa[IFA_ADDRESS]) {
		char paddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_ADDRESS]), paddr, sizeof(paddr));
		rec_dbg("    IFA_ADDRESS(%hu): %s",
			RTA_ALIGN(ifa[IFA_ADDRESS]->rta_len), paddr);
	}

	/* debug local address */
	if(ifa[IFA_LOCAL]) {
		char laddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_LOCAL]), laddr, sizeof(laddr));
		rec_dbg("    IFA_LOCAL(%hu): %s",
			RTA_ALIGN(ifa[IFA_LOCAL]->rta_len), laddr);
	}

	/* debug interface label */
	if(ifa[IFA_LABEL])
		rec_dbg("    IFA_LABEL(%hu): %s",
			RTA_ALIGN(ifa[IFA_LABEL]->rta_len), (char *)RTA_DATA(ifa[IFA_LABEL]));

	/* debug interface broadcast address */
	if(ifa[IFA_BROADCAST]) {
		char baddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_BROADCAST]), baddr, sizeof(baddr));
		rec_dbg("    IFA_BROADCAST(%hu): %s",
			RTA_ALIGN(ifa[IFA_BROADCAST]->rta_len), baddr);
	}

	/* debug interface anycast address */
	if(ifa[IFA_ANYCAST]) {
		char aaddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_ANYCAST]), aaddr, sizeof(aaddr));
		rec_dbg("    IFA_ANYCAST(%hu): %s",
			RTA_ALIGN(ifa[IFA_ANYCAST]->rta_len), aaddr);
	}

	/* debug interface cache information */
	if(ifa[IFA_CACHEINFO]) {
		struct ifa_cacheinfo *ifac = (struct ifa_cacheinfo *)RTA_DATA(ifa[IFA_CACHEINFO]);
		rec_dbg("    IFA_CACHEINFO(%hu):",
			RTA_ALIGN(ifa[IFA_CACHEINFO]->rta_len));
		rec_dbg("        [ ifa_cacheinfo(%d) ]",
			NLMSG_ALIGN(sizeof(struct ifa_cacheinfo)));
		rec_dbg("            ifa_prefered(%d): 0x%x",
			sizeof(ifac->ifa_prefered), ifac->ifa_prefered);
		rec_dbg("            ifa_valid(%d): 0x%x",
			sizeof(ifac->ifa_valid), ifac->ifa_valid);
		rec_dbg("            cstamp(%d): %u",
			sizeof(ifac->cstamp), ifac->cstamp);
		rec_dbg("            tstamp(%d): %u",
			sizeof(ifac->tstamp), ifac->tstamp);
	}
	
	/* debug interface multicast address */
	if(ifa[IFA_MULTICAST]) {
		char maddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_MULTICAST]), maddr, sizeof(maddr));
		rec_dbg("    IFA_ANYCAST(%hu): %s",
			RTA_ALIGN(ifa[IFA_MULTICAST]->rta_len), maddr);
	}

	rec_dbg("");

	return;
}

/*
 * convert interface address flags from number to string
 */
void debug_n2s_ifa_flags(int flags, char *flags_list)
{
#define _IFA_FLAGS(f) \
	if(flags & IFA_F_##f) \
		(flags &= ~IFA_F_##f) ? strcat(flags_list, #f ",") : strcat(flags_list, #f);
	_IFA_FLAGS(SECONDARY);
	_IFA_FLAGS(TEMPORARY);
	_IFA_FLAGS(NODAD);
	_IFA_FLAGS(OPTIMISTIC);
	_IFA_FLAGS(DADFAILED);
	_IFA_FLAGS(HOMEADDRESS);
	_IFA_FLAGS(DEPRECATED);
	_IFA_FLAGS(TENTATIVE);
	_IFA_FLAGS(PERMANENT);
#undef _IFA_FLAGS
}

/*
 * convert interface address scope from number to string
 */
const char *debug_n2s_ifa_scope(int scope)
{
#define _IFA_SCOPE(s) \
	if(scope == RT_SCOPE_##s) \
		return #s;
	_IFA_SCOPE(UNIVERSE);
	_IFA_SCOPE(SITE);
	_IFA_SCOPE(LINK);
	_IFA_SCOPE(HOST);
	_IFA_SCOPE(NOWHERE);
#undef _IFA_SCOPE
}
