#include "nield.h"
#include "rtnetlink.h"

#define HIST_MAX 8

extern int optflag;

/* top of interface list and history */
static LIST_HEAD(lhead);
static LIST_HEAD(hhead);

/*
 * search an entry in an interface list
 */
static inline struct iflist_entry *search_iflist(unsigned index)
{
	struct list_head *l;

	list_for_each(l, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index)
			return(e);
	}

	return(NULL);
}

/*
 * add an entry into an interface list
 */ 
static inline void add_iflist_entry(struct list_head *l)
{
	INIT_LIST_HEAD(l);
	list_add(l, &lhead);
}

/*
 * [not in use]
 * delete an entry in an interface list
 */ 
static inline void del_iflist_entry(unsigned index)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index) {
			list_del(l);
			free(e);
		}
	}
}

/*
 * move an entry in an interface list
 */
static inline void move_iflist_entry(unsigned index)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index) {
			list_move(l, &hhead);
			break;
		}
	}
}

/*
 * delete an entry in an interface history
 */
static inline void del_ifhist_entry(void)
{
	struct list_head *l, *n;
	int i = 0;

	list_for_each_safe(l, n, &hhead) {
		struct iflist_entry *e;

		if(++i > HIST_MAX) {
			e = list_entry(l, struct iflist_entry, list);
			list_del(l);
			free(e);
			break;
		}
	}
}

/*
 * search an entry in an interface history
 */ 
static inline struct iflist_entry *search_ifhist(unsigned index)
{
	struct list_head *l;

	list_for_each(l, &hhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index)
			return(e);
	}

	return(NULL);
}

/*
 * convert an interface index to name in an interface list
 */
char *if_indextoname_from_iflist(unsigned index, char *ifname)
{
	struct iflist_entry *e = search_iflist(index);

	if(!e)
		return(NULL);

	strncpy(ifname, e->name, sizeof(e->name));

	return(ifname);
}

/*
 * convert an interface index to name in an inetrface history
 */ 
char *if_indextoname_from_ifhist(unsigned index, char *ifname)
{
	struct iflist_entry *e = search_ifhist(index);

	if(!e)
		return(NULL);

	strncpy(ifname, e->name, sizeof(e->name));

	return(ifname);
}

/*
 * get an interface type in an interface list
 */ 
unsigned short get_type_from_iflist(unsigned index)
{
	struct iflist_entry *e = search_iflist(index);

	if(!e)
		return(0);

	return(e->type);
}

/*
 * get an interface type in an interface history
 */
unsigned short get_type_from_ifhist(unsigned index)
{
	struct iflist_entry *e = search_ifhist(index);

	if(!e)
		return(0);

	return(e->type);
}

/*
 * [not in use]
 * print an interface list
 */ 
void print_iflist(void)
{
	struct list_head *l;

	list_for_each(l, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);

		printf("%d ", e->index);
		printf("IFLA_IFNAME: %s, ", e->name);
		printf("IFLA_ADDRESS: %s, ", e->addr);
		printf("IFLA_BROADCAST: %s, ", e->brd);
		printf("IFLA_MTU: %hu\n", e->mtu);
	}
}

/*
 * create an interface list
 */ 
int create_iflist(struct msghdr *msg)
{
	struct nlmsghdr *nlh;
	int nlh_len;
	struct ifinfomsg *ifim;
	int ifim_len;
	struct rtattr *ifla[__IFLA_MAX];
	struct iflist_entry *ifle;

	/* get netlink message header */
	nlh = msg->msg_iov->iov_base;
	nlh_len = msg->msg_iov->iov_len;

	/* parse netlink message header */
	for( ; NLMSG_OK(nlh, nlh_len); nlh = NLMSG_NEXT(nlh, nlh_len)) {
		/* whether netlink message header ends or not */
		if(nlh->nlmsg_type == NLMSG_DONE)
			return(1);

		/* get interface infomation message */
		ifim = (struct ifinfomsg *)NLMSG_DATA(nlh);
		ifim_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);

		/* parse interface infomation attributes */
		parse_ifinfo(ifla, nlh);

		/* debug */
		if(optflag & F_DEBUG) {
			debug_nlmsg(nlh);
			debug_ifimsg(ifim, ifla, ifim_len);
		}

		/* check bridge interface */
		if(ifim->ifi_family == PF_BRIDGE) {
			ifle = search_iflist(ifim->ifi_index);
			if(ifle)
				ifle->br_attached = 1;
		}

		/* check protocol family(PF_UNSPEC only) */
		if(ifim->ifi_family != PF_UNSPEC)
			continue;

		/* create interface list */
		ifle = (struct iflist_entry *)malloc(sizeof(struct iflist_entry));
		if(ifle == NULL) {
			rec_log("[%-12s] %s: malloc(): failed", "ERROR", __func__);
			return(-1);
		}
		memset(ifle, 0, sizeof(struct iflist_entry));
		INIT_LIST_HEAD(&(ifle->list));

		/* get interface name */
		ifle->index = ifim->ifi_index;
		if(ifla[IFLA_IFNAME])
			memcpy(ifle->name, RTA_DATA(ifla[IFLA_IFNAME]), sizeof(ifle->name));

		/* get interface address */
		if(ifla[IFLA_ADDRESS]) {
			char addrstr[INET6_ADDRSTRLEN+1] = "";
			inet_ntop_if_addr(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
				RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
			memcpy(ifle->addr, addrstr, sizeof(addrstr));
		}

		/* get broadcast address */
		if(ifla[IFLA_BROADCAST]) {
			char brdstr[INET6_ADDRSTRLEN+1] = "";
			inet_ntop_if_addr(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
				RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
			memcpy(ifle->brd, brdstr, sizeof(brdstr));
		}

		/* get interface broadcast address */
		if(ifla[IFLA_BROADCAST])
			memcpy(ifle->brd, RTA_DATA(ifla[IFLA_BROADCAST]), sizeof(ifle->brd));

		ifle->flags = ifim->ifi_flags;
		ifle->type = ifim->ifi_type;
		ifle->vid = 0;

		/* get interface information */
		if(ifla[IFLA_LINKINFO]) {
			struct rtattr *linkinfo[__IFLA_INFO_MAX];
			parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
			if (linkinfo[IFLA_INFO_KIND]) {
				memcpy(ifle->kind, (char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]),
					sizeof(ifle->kind));
				if(!strcmp(ifle->kind, "vlan")) {
					struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
					parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
					if(vlaninfo[IFLA_VLAN_ID])
						ifle->vid = *(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]);
				}
			}
		}

		/* get interface MTU */
		if(ifla[IFLA_MTU])
			ifle->mtu = *(unsigned short *)RTA_DATA(ifla[IFLA_MTU]);

		/* get master interface */
		if(ifla[IFLA_MASTER])
			ifle->master = *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]);

		/* add interface list */
		list_add(&(ifle->list), &lhead);
	}

	return(0);
}

/*
 * parse interface information messages
 */
int parse_ifimsg(struct nlmsghdr *nlh)
{
	struct ifinfomsg *ifim = (struct ifinfomsg *)NLMSG_DATA(nlh);
	int ifim_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);
	struct rtattr *ifla[__IFLA_MAX];
	struct iflist_entry *ifle_tmp, *ifle;
	char msg_type[MAX_STR_SIZE] = "", msg[MAX_MSG_SIZE] = "";
	char *mp = msg;
	char ifname_master[IFNAMSIZ] = "";

	/* parse interface infomation message attributes */
	parse_ifinfo(ifla, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ifimsg(ifim, ifla, ifim_len);
	}

	/* set message type */
	strcpy(msg_type, "INTERFACE");

	/* create new interface list entry */
	ifle_tmp = malloc(sizeof(struct iflist_entry));
	if(!ifle_tmp) {
		rec_log("[%-12s] %s: malloc() failed", "ERROR", __func__);
		return(1);
	}
	memset(ifle_tmp, 0, sizeof(struct iflist_entry));
	INIT_LIST_HEAD(&(ifle_tmp->list));

	ifle_tmp->index = ifim->ifi_index;
	ifle_tmp->flags = ifim->ifi_flags;
	ifle_tmp->type = ifim->ifi_type;

	/* get interface name */
	if(ifla[IFLA_IFNAME]) {
		strncpy(ifle_tmp->name, RTA_DATA(ifla[IFLA_IFNAME]), IFNAMSIZ);
		mp += sprintf(mp, "if=%s ", ifle_tmp->name);
	}

	/* get physical interface */
	if(ifla[IFLA_LINK]) {
		unsigned int iflinkindex = *(unsigned *)RTA_DATA(ifla[IFLA_LINK]);
		char iflinkname[IFNAMSIZ] = "";
		if(ifle_tmp->index != iflinkindex) {
			if(!if_indextoname(iflinkindex, iflinkname))
				if(!if_indextoname_from_iflist(iflinkindex, iflinkname))
					if_indextoname_from_ifhist(iflinkindex, iflinkname);
			mp += sprintf(mp, "link=%s ", iflinkname);
		}
	}

	/* get interface address */
	if(ifla[IFLA_ADDRESS]) {
		char addrstr[INET6_ADDRSTRLEN+1] = "";
		inet_ntop_if_addr(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
			RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
		memcpy(ifle_tmp->addr, addrstr, sizeof(addrstr));
		switch(ifle_tmp->type) {
			case ARPHRD_TUNNEL:
			case ARPHRD_IPGRE:
			case ARPHRD_SIT:
			case ARPHRD_TUNNEL6:
				mp += sprintf(mp, "local=%s ", ifle_tmp->addr);
				break;
			default:
				mp += sprintf(mp, "lladdr=%s ", ifle_tmp->addr);
		}
	}

	/* get broadcast address */
	if(ifla[IFLA_BROADCAST]) {
		char brdstr[INET6_ADDRSTRLEN+1] = "";
		inet_ntop_if_addr(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
			RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
		memcpy(ifle_tmp->brd, brdstr, sizeof(brdstr));
		switch(ifle_tmp->type) {
			case ARPHRD_TUNNEL:
			case ARPHRD_IPGRE:
			case ARPHRD_SIT:
			case ARPHRD_TUNNEL6:
				mp += sprintf(mp, "remote=%s ", ifle_tmp->brd);
				break;
		}
	}

	/* get interface MTU */
	if(ifla[IFLA_MTU]) {
		ifle_tmp->mtu = *(unsigned *)RTA_DATA(ifla[IFLA_MTU]);
		mp += sprintf(mp, "mtu=%hu ", ifle_tmp->mtu);
	}

	/* get interface information */
	if(ifla[IFLA_LINKINFO]) {
		struct rtattr *linkinfo[__IFLA_INFO_MAX];
		parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
		if(linkinfo[IFLA_INFO_KIND]) {
			memcpy(ifle_tmp->kind, (char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]),
				sizeof(ifle_tmp->kind));
			mp += sprintf(mp, "kind=%s ", ifle_tmp->kind);
			if(!strcmp(ifle_tmp->kind, "vlan")) {
				struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
				parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);

				if(vlaninfo[IFLA_VLAN_ID]) {
					ifle_tmp->vid = *(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]);
					mp += sprintf(mp, "vlan=%hu ", ifle_tmp->vid);
				}
			} else if(!strcmp(ifle_tmp->kind, "gretap")) {
				struct rtattr *greinfo[__IFLA_GRE_MAX];
				parse_greinfo(greinfo, linkinfo[IFLA_INFO_DATA]);

				if(greinfo[IFLA_GRE_LOCAL]) {
					char gre_laddr[INET6_ADDRSTRLEN+1] = "";

					inet_ntop(AF_INET, RTA_DATA(greinfo[IFLA_GRE_LOCAL]),
						gre_laddr, sizeof(gre_laddr));
					mp += sprintf(mp, "local=%s ", gre_laddr);
				}

				if(greinfo[IFLA_GRE_REMOTE]) {
					char gre_raddr[INET6_ADDRSTRLEN+1] = "";

					inet_ntop(AF_INET, RTA_DATA(greinfo[IFLA_GRE_REMOTE]),
						gre_raddr, sizeof(gre_raddr));
					mp += sprintf(mp, "remote=%s ", gre_raddr);
				}
			}
		}
	}

	/* get master interface */
	unsigned int ifindex_master = 0;
	if(ifla[IFLA_MASTER]) {
		ifindex_master = *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]);
		ifle_tmp->master = ifindex_master;
		if(!if_indextoname(ifindex_master, ifname_master))
			if(!if_indextoname_from_iflist(ifindex_master, ifname_master))
				if_indextoname_from_ifhist(ifindex_master, ifname_master);
		mp += sprintf(mp, "master=%s ", ifname_master);
	}

	/* check protocol family(PF_BRIDGE only) & nlmsg_type */
	if(ifim->ifi_family == PF_BRIDGE) {
		/* workaround: "ovs-vsctl add-br" command */
		if(ifle_tmp->index != ifindex_master) {
			if(nlh->nlmsg_type == RTM_NEWLINK) {
				ifle = search_iflist(ifle_tmp->index);
				if(ifle && !ifle->br_attached) {
					ifle->br_attached = 1;
					rec_log("[%-12s] interface %s attached to the ethernet bridge %s",
						msg_type, ifle->name, ifname_master);
				}
			} else if(nlh->nlmsg_type == RTM_DELLINK) {
				ifle = search_iflist(ifle_tmp->index);
				if(ifle && ifle->br_attached) {
					ifle->br_attached = 0;
					rec_log("[%-12s] interface %s detached from the ethernet bridge %s",
						msg_type, ifle->name, ifname_master);
				}
			} else {
				rec_log("[%-12s] %s: unkonwn nlmsg_type: %d",
					"ERROR", __func__, nlh->nlmsg_type);
			}
		}
	}

	/* check protocol family(PF_UNSPEC only) */
	if(ifim->ifi_family != PF_UNSPEC)
		return(0);

	/* check nlmsg_type */
	if(nlh->nlmsg_type == RTM_NEWLINK) {
		/* search interface list entry */
		ifle = search_iflist(ifle_tmp->index);
		if(ifle) {
			/* check bonding interface */
			if((ifle->flags & IFF_SLAVE) && !(ifle_tmp->flags & IFF_SLAVE)) {
				if(!if_indextoname(ifle->master, ifname_master))
					if(!if_indextoname_from_iflist(ifle->master, ifname_master))
						if_indextoname_from_ifhist(ifle->master, ifname_master);
				rec_log("[%-12s] interface %s detached from the bonding interface %s",
					msg_type, ifle_tmp->name, ifname_master);
			} else if(!(ifle->flags & IFF_SLAVE) && (ifle_tmp->flags & IFF_SLAVE))
				rec_log("[%-12s] interface %s attached to the bonding interface %s",
					msg_type, ifle_tmp->name, ifname_master);

			/* check administrative status */
			if((ifle->flags & IFF_UP) && !(ifle_tmp->flags & IFF_UP))
				rec_log("[%-12s] interface %s state changed to disabled",
					msg_type, ifle_tmp->name);
			else if(!(ifle->flags & IFF_UP) && (ifle_tmp->flags & IFF_UP))
				rec_log("[%-12s] interface %s state changed to enabled",
					msg_type, ifle_tmp->name);

			/* check operational status */
			if((ifle->flags & IFF_RUNNING) && !(ifle_tmp->flags & IFF_RUNNING))
				rec_log("[%-12s] interface %s state changed to down",
					msg_type, ifle_tmp->name);
			else if(!(ifle->flags & IFF_RUNNING) && (ifle_tmp->flags & IFF_RUNNING))
				rec_log("[%-12s] interface %s state changed to up",
					msg_type, ifle_tmp->name);

			ifle->flags = ifle_tmp->flags;

			/* check interface name */
			if(ifla[IFLA_IFNAME]) {
				if(strncmp(ifle->name, ifle_tmp->name, IFNAMSIZ)) {
					rec_log("[%-12s] interface name changed from %s to %s",
						msg_type, ifle->name, ifle_tmp->name);
					strncpy(ifle->name, ifle_tmp->name, IFNAMSIZ);
				}
			}

			/* check interface address */
			if(ifla[IFLA_ADDRESS]) {
				if(memcmp(ifle->addr, ifle_tmp->addr, sizeof(ifle->addr))) {
					switch(ifle_tmp->type) {
						case ARPHRD_TUNNEL:
						case ARPHRD_IPGRE:
						case ARPHRD_SIT:
						case ARPHRD_TUNNEL6:
							rec_log("[%-12s] interface %s local address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->addr, ifle_tmp->addr);
							break;
						default:
							rec_log("[%-12s] interface %s link layer address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->addr, ifle_tmp->addr);
					}
					memcpy(ifle->addr, ifle_tmp->addr, sizeof(ifle->addr));
				}
			}

			/* check broadcast address */
			if(ifla[IFLA_BROADCAST]) {
				if(memcmp(ifle->brd, ifle_tmp->brd, sizeof(ifle->brd))) {
					switch(ifle_tmp->type) {
						case ARPHRD_TUNNEL:
						case ARPHRD_IPGRE:
						case ARPHRD_SIT:
						case ARPHRD_TUNNEL6:
							rec_log("[%-12s] interface %s remote address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->brd, ifle_tmp->brd);
							break;
					}
					memcpy(ifle->brd, ifle_tmp->brd, sizeof(ifle->brd));
				}
			}

			ifle->type = ifle_tmp->type;

			/* check interface MTU */
			if(ifla[IFLA_MTU]) {
				if(ifle->mtu != ifle_tmp->mtu) {
					rec_log("[%-12s] interface %s mtu changed from %hu to %hu",
						msg_type, ifle->name, ifle->mtu, ifle_tmp->mtu);
					ifle->mtu = ifle_tmp->mtu;
				}
			} 

			/* check interface vlan id */
			if(ifle->vid != ifle_tmp->vid) {
				rec_log("[%-12s] interface %s vlan id changed from %hu to %hu",
					msg_type, ifle->name, ifle->vid, ifle_tmp->vid);
				ifle->vid = ifle_tmp->vid;
			}

			/* check master interface */
			if(ifle->master != ifle_tmp->master) {
				ifle->master = ifle_tmp->master;
			}

			free(ifle_tmp);
		} else {
			/* add interface list entry*/
			add_iflist_entry(&(ifle_tmp->list));

			/* check interface state */
			char state[20] = "";

			/* check administrative state */
			(ifle_tmp->flags & IFF_UP) ? strcpy(state, "enabled,") : strcpy(state, "disabled,");

			/* check operational state */
			(ifle_tmp->flags & IFF_RUNNING) ? strcat(state, "linkup") : strcat(state, "linkdown");

			rec_log("[%-12s] interface added: %sstate=%s", msg_type, msg, state);
		}
	} else if(nlh->nlmsg_type == RTM_DELLINK) {
		/* move entry from interface list to interface history */
		move_iflist_entry(ifle_tmp->index);
		del_ifhist_entry();

		/* delete interface information list */
		//del_iflist_entry(ifle_tmp->index);

		/* check interface state */
		char state[20] = "";

		/* check administrative state */
		(ifle_tmp->flags & IFF_UP) ? strcpy(state, "enabled,") : strcpy(state, "disabled,");

		/* check operational state */
		(ifle_tmp->flags & IFF_RUNNING) ? strcat(state, "linkup") : strcat(state, "linkdown");

		rec_log("[%-12s] interface deleted: %sstate=%s", msg_type, msg, state);

		free(ifle_tmp);
	} else {
		rec_log("[%-12s] unkonwn nlmsg_type: %d", "ERROR", nlh->nlmsg_type);

		free(ifle_tmp);
	}

	return(0);
}

/*
 * convert interface address from binary to text
 */ 
void inet_ntop_if_addr(unsigned short type, unsigned char *addr, int addrlen, char *addrstr, int addrstrlen)
{
	int i;

	switch(type) {
		case ARPHRD_TUNNEL:
		case ARPHRD_IPGRE:
		case ARPHRD_SIT:
			inet_ntop(AF_INET, addr, addrstr, addrstrlen);
			break;
		case ARPHRD_TUNNEL6:
			inet_ntop(AF_INET6, addr, addrstr, addrstrlen);
			break;
		default:
			for(i = 0; i < addrlen; i++)
				addrstr += sprintf(addrstr, "%.2x:", addr[i]);
			addrstr--;
			*addrstr = '\0';
	}
}

/*
 * debug interface information messages
 */ 
void debug_ifimsg(struct ifinfomsg *ifim, struct rtattr *ifla[], int ifim_len)
{
	/* debug ifinfomsg */
	char ifname[IFNAMSIZ] = "";
	char ifi_flags_list[MAX_STR_SIZE] = "";

	rec_dbg("*********************************************************************");
	rec_dbg("[ ifinfomsg(%d) ]",
		NLMSG_ALIGN(sizeof(struct ifinfomsg)));
	rec_dbg("    ifi_family(%d): %d(%s)",
		sizeof(ifim->ifi_family), ifim->ifi_family, debug_n2s_af_type(ifim->ifi_family));
	rec_dbg("    __ifi_pad(%d): %d",
		sizeof(ifim->__ifi_pad), ifim->__ifi_pad);
	rec_dbg("    ifi_type(%d): %hu(%s)",
		sizeof(ifim->ifi_type), ifim->ifi_type, debug_n2s_ifi_type(ifim->ifi_type));
	if(!if_indextoname(ifim->ifi_index, ifname))
		if(!if_indextoname_from_iflist(ifim->ifi_index, ifname))
			if_indextoname_from_ifhist(ifim->ifi_index, ifname);
	rec_dbg("    ifi_index(%d): %d(%s)",
		sizeof(ifim->ifi_index), ifim->ifi_index, ifname);
	debug_n2s_ifi_flags(ifim->ifi_flags, ifi_flags_list);
	rec_dbg("    ifi_flags(%d): 0x%.8x(%s)",
		sizeof(ifim->ifi_flags), ifim->ifi_flags, ifi_flags_list);
	rec_dbg("    ifi_change(%d): %u",
		sizeof(ifim->ifi_change), ifim->ifi_change);

	/* debug interface link attributes */
	rec_dbg("*********************************************************************");
	rec_dbg("[ ifinfomsg attributes(%d) ]",
		NLMSG_ALIGN(ifim_len - NLMSG_ALIGN(sizeof(struct ifinfomsg))));

	/* debug interface address */
	if(ifla[IFLA_ADDRESS]) {
		char addrstr[INET6_ADDRSTRLEN+1] = "";
		inet_ntop_if_addr(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
			RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
		rec_dbg("    IFLA_ADDRESS(%hu): %s",
			RTA_ALIGN(ifla[IFLA_ADDRESS]->rta_len), addrstr);
	}

	/* debug interface broadcast address */
	if(ifla[IFLA_BROADCAST]) {
		char brdstr[INET6_ADDRSTRLEN+1] = "";
		inet_ntop_if_addr(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
			RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
		rec_dbg("    IFLA_BROADCAST(%hu): %s",
			RTA_ALIGN(ifla[IFLA_BROADCAST]->rta_len), brdstr);
	}

	/* debug interface name */
	if(ifla[IFLA_IFNAME])
		rec_dbg("    IFLA_IFNAME(%hu): %s",
			RTA_ALIGN(ifla[IFLA_IFNAME]->rta_len),
			(char *)RTA_DATA(ifla[IFLA_IFNAME]));

	/* debug interface MTU */
	if(ifla[IFLA_MTU])
		rec_dbg("    IFLA_MTU(%hu): %hu",
			RTA_ALIGN(ifla[IFLA_MTU]->rta_len),
			*(unsigned short *)RTA_DATA(ifla[IFLA_MTU]));

	/* debug physical interface index */
	if(ifla[IFLA_LINK]) {
		unsigned int iflinkindex = *(unsigned *)RTA_DATA(ifla[IFLA_LINK]);
		char iflinkname[IFNAMSIZ] = "";
		if(!if_indextoname(iflinkindex, iflinkname))
			if(!if_indextoname_from_iflist(iflinkindex, iflinkname))
				if_indextoname_from_ifhist(iflinkindex, iflinkname);
		rec_dbg("    IFLA_LINK(%hu): %u(%s)",
			RTA_ALIGN(ifla[IFLA_LINK]->rta_len), iflinkindex, iflinkname);
	}

	/* debug interface queue discipline */
	if(ifla[IFLA_QDISC])
		rec_dbg("    IFLA_QDISC(%hu): %s",
			RTA_ALIGN(ifla[IFLA_QDISC]->rta_len), (char *)RTA_DATA(ifla[IFLA_QDISC]));

	/* debug interface statistics */
	if(ifla[IFLA_STATS]) {
		struct rtnl_link_stats *stats = (struct rtnl_link_stats *)RTA_DATA(ifla[IFLA_STATS]);
		rec_dbg("    IFLA_STATS(%hu):",
			RTA_ALIGN(ifla[IFLA_STATS]->rta_len));
		rec_dbg("        [ rtnl_link_stats(%d) ]",
			NLMSG_ALIGN(sizeof(struct rtnl_link_stats)));
		rec_dbg("            rx_packets(%d): %u",
			sizeof(stats->rx_packets), stats->rx_packets);
		rec_dbg("            tx_packets(%d): %u",
			sizeof(stats->tx_packets), stats->tx_packets);
		rec_dbg("            rx_bytes(%d): %u",
			sizeof(stats->rx_bytes), stats->rx_bytes);
		rec_dbg("            tx_bytes(%d): %u",
			sizeof(stats->tx_bytes), stats->tx_bytes);
		rec_dbg("            rx_errors(%d): %u",
			sizeof(stats->rx_errors), stats->rx_errors);
		rec_dbg("            tx_errors(%d): %u",
			sizeof(stats->tx_errors), stats->tx_errors);
		rec_dbg("            rx_dropped(%d): %u",
			sizeof(stats->rx_dropped), stats->rx_dropped);
		rec_dbg("            tx_dropped(%d): %u",
			sizeof(stats->tx_dropped), stats->tx_dropped);
		rec_dbg("            multicast(%d): %u",
			sizeof(stats->multicast), stats->multicast);
		rec_dbg("            collisions(%d): %u",
			sizeof(stats->collisions), stats->collisions);
		rec_dbg("            rx_length_errors(%d): %u",
			sizeof(stats->rx_length_errors), stats->rx_length_errors);
		rec_dbg("            rx_over_errors(%d): %u",
			sizeof(stats->rx_over_errors), stats->rx_over_errors);
		rec_dbg("            rx_crc_errors(%d): %u",
			sizeof(stats->rx_crc_errors), stats->rx_crc_errors);
		rec_dbg("            rx_frame_errors(%d): %u",
			sizeof(stats->rx_frame_errors), stats->rx_frame_errors);
		rec_dbg("            rx_fifo_errors(%d): %u",
			sizeof(stats->rx_fifo_errors), stats->rx_fifo_errors);
		rec_dbg("            rx_missed_errors(%d): %u",
			sizeof(stats->rx_missed_errors), stats->rx_missed_errors);
		rec_dbg("            tx_aborted_errors(%d): %u",
			sizeof(stats->tx_aborted_errors), stats->tx_aborted_errors);
		rec_dbg("            tx_carrier_errors(%d): %u",
			sizeof(stats->tx_carrier_errors), stats->tx_carrier_errors);
		rec_dbg("            tx_fifo_errors(%d): %u",
			sizeof(stats->tx_fifo_errors), stats->tx_fifo_errors);
		rec_dbg("            tx_heartbeat_errors(%d): %u",
			sizeof(stats->tx_heartbeat_errors), stats->tx_heartbeat_errors);
		rec_dbg("            tx_window_errors(%d): %u",
			sizeof(stats->tx_window_errors), stats->tx_window_errors);
		rec_dbg("            rx_compressed(%d): %u",
			sizeof(stats->rx_compressed), stats->rx_compressed);
		rec_dbg("            tx_compressed(%d): %u",
			sizeof(stats->tx_compressed), stats->tx_compressed);
	}

	/* unknown attribute */
	if(ifla[IFLA_COST])
		rec_dbg("    IFLA_COST(%hu): %u",
			RTA_ALIGN(ifla[IFLA_COST]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_COST]));

	/* unknown attribute */
	if(ifla[IFLA_PRIORITY])
		rec_dbg("    IFLA_PRIORITY(%hu): %u",
			RTA_ALIGN(ifla[IFLA_PRIORITY]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_PRIORITY]));

	/* debug master(parent) interface index */
	if(ifla[IFLA_MASTER]) {
		unsigned int ifindex_master = *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]);
		char ifname_master[IFNAMSIZ] = "";
		if(!if_indextoname(ifindex_master, ifname_master))
			if(!if_indextoname_from_iflist(ifindex_master, ifname_master))
				if_indextoname_from_ifhist(ifindex_master, ifname_master);
		rec_dbg("    IFLA_MASTER(%hu): %u(%s)",
			RTA_ALIGN(ifla[IFLA_MASTER]->rta_len), ifindex_master, ifname_master);
	}

	/* debug wireless interface information(struct iw_event) */
	if(ifla[IFLA_WIRELESS])
		rec_dbg("    IFLA_WIRELESS(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_WIRELESS]->rta_len));

	/* debug interface protocol information */
	if(ifla[IFLA_PROTINFO])
		rec_dbg("    IFLA_PROTINFO(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_PROTINFO]->rta_len));

	/* debug interafce transmit queue length */
	if(ifla[IFLA_TXQLEN])
		rec_dbg("    IFLA_TXQLEN(%hu): %u",
			RTA_ALIGN(ifla[IFLA_TXQLEN]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_TXQLEN]));

	/* debug interface memory mapping */
	if(ifla[IFLA_MAP]) {
		struct rtnl_link_ifmap *ifmap = (struct rtnl_link_ifmap *)RTA_DATA(ifla[IFLA_MAP]);
		unsigned char *mem = (unsigned char *)&(ifmap->mem_start);
		rec_dbg("    IFLA_MAP(%hu):",
			RTA_ALIGN(ifla[IFLA_MAP]->rta_len));
		rec_dbg("        [ rtnl_link_ifmap(%d) ]",
			NLMSG_ALIGN(sizeof(struct rtnl_link_ifmap)));
		rec_dbg("            mem_start(%d): 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			sizeof(ifmap->mem_start),
			mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		mem = (unsigned char *)&(ifmap->mem_end);
		rec_dbg("            mem_end(%d): 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			sizeof(ifmap->mem_end),
			mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		mem = (unsigned char *)&(ifmap->base_addr);
		rec_dbg("            base_addr(%d): 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			sizeof(ifmap->base_addr),
			mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		rec_dbg("            irq(%d): %hu",
			sizeof(ifmap->irq), ifmap->irq);
		rec_dbg("            dma(%d): 0x%.2x",
			sizeof(ifmap->dma), ifmap->dma);
		rec_dbg("            port(%d): 0x%.2x",
			sizeof(ifmap->port), ifmap->port);
	}

	/* unknown attribute */
	if(ifla[IFLA_WEIGHT])
		rec_dbg("    IFLA_WEIGHT(%hu): %u",
			RTA_ALIGN(ifla[IFLA_WEIGHT]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_WEIGHT]));

	/* debug interface operational state(RFC 2863) */
	if(ifla[IFLA_OPERSTATE])
		rec_dbg("    IFLA_OPERSTATE(%hu): %d(%s)",
			RTA_ALIGN(ifla[IFLA_OPERSTATE]->rta_len),
			*(int *)RTA_DATA(ifla[IFLA_OPERSTATE]),
			debug_n2s_if_oper_state(*(int *)RTA_DATA(ifla[IFLA_OPERSTATE])));

	/* debug interface link mode */
	if(ifla[IFLA_LINKMODE])
		rec_dbg("    IFLA_LINKMODE(%hu): %d(%s)",
			RTA_ALIGN(ifla[IFLA_LINKMODE]->rta_len),
			*(int *)RTA_DATA(ifla[IFLA_LINKMODE]),
			debug_n2s_ifla_linkmode(*(int *)RTA_DATA(ifla[IFLA_LINKMODE])));

	/*
	 * debug interface link information
	 *
	 * IFLA_LINKINFO was committed in kernel 2.6.23.
	 *
	 * commit 38f7b870d4a6a5d3ec21557e849620cb7d032965
	 * Author: Patrick McHardy <kaber@trash.net>
	 * Date:   Wed Jun 13 12:03:51 2007 -0700
	 *
	 *     [RTNETLINK]: Link creation API
	 */
	if(ifla[IFLA_LINKINFO]) {
		struct rtattr *linkinfo[__IFLA_INFO_MAX];
		parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
		rec_dbg("    IFLA_LINKINFO(%hu):",
			RTA_ALIGN(ifla[IFLA_LINKINFO]->rta_len));
		if(linkinfo[IFLA_INFO_KIND]) {
			rec_dbg("        IFLA_INFO_KIND(%hu): %s",
				RTA_ALIGN(linkinfo[IFLA_INFO_KIND]->rta_len),
				(char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]));
			if(linkinfo[IFLA_INFO_DATA])
				rec_dbg("        IFLA_INFO_DATA(%hu):",
					RTA_ALIGN(linkinfo[IFLA_INFO_DATA]->rta_len));
			if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "vlan")) {
				struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
				parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
				debug_vlaninfo(vlaninfo);
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "gre")) {
				struct rtattr *greinfo[__IFLA_GRE_MAX];
				parse_greinfo(greinfo, linkinfo[IFLA_INFO_DATA]);
				debug_greinfo(greinfo);
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "gretap")) {
				struct rtattr *greinfo[__IFLA_GRE_MAX];
				parse_greinfo(greinfo, linkinfo[IFLA_INFO_DATA]);
				debug_greinfo(greinfo);
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "tun")) {
				/* ignore */
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "bond")) {
				/* ignore */
			}
		}

		if(linkinfo[IFLA_INFO_XSTATS]) {
			rec_dbg("        IFLA_INFO_XSTATS(%hu): -- ignored --",
				RTA_ALIGN(linkinfo[IFLA_INFO_XSTATS]->rta_len));
		}
	}

	/* 
	 * debug pid of process in network name space
	 *
	 * IFLA_NET_NS_PID was committed in kernel 2.6.24.
	 *
	 * commit d8a5ec672768c3cf4d51d7a63fc071520afa1617
	 * Author: Eric W. Biederman <ebiederm@xmission.com>
	 * Date:   Wed Sep 12 13:57:04 2007 +0200
	 *
	 *     [NET]: netlink support for moving devices between network namespaces.
	 */
	if(ifla[IFLA_NET_NS_PID])
		rec_dbg("    IFLA_NET_NS_PID(%hu): %u",
			RTA_ALIGN(ifla[IFLA_NET_NS_PID]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_NET_NS_PID]));

	/*
	 * debug interface alias name
	 *
	 * IFLA_IFALIAS was committed in kernel 2.6.28.
	 *
	 * commit 0b815a1a6d43ab498674b8430c8c35ab08487a16
	 * Author: Stephen Hemminger <shemminger@vyatta.com>
	 * Date:   Mon Sep 22 21:28:11 2008 -0700
	 *
	 *     net: network device name ifalias support
	 */ 
	if(ifla[IFLA_IFALIAS])
		rec_dbg("    IFLA_IFALIAS(%hu): %s",
			RTA_ALIGN(ifla[IFLA_IFALIAS]->rta_len),
			(char *)RTA_DATA(ifla[IFLA_IFALIAS]));

	/*
	 * debug number of VFs
	 *
	 * IFLA_NUM_VF was committed in kernel 2.6.34.
	 *
	 * commit b280da8d54b8d82b52f368a8703b7ada6c1744d5
	 * Author: Williams, Mitch A <mitch.a.williams@intel.com>
	 * Date:   Wed Feb 10 01:43:24 2010 +0000
	 *
	 *     if_link: Add SR-IOV configuration methods
	 */
	if(ifla[IFLA_NUM_VF])
		rec_dbg("    IFLA_NUM_VF(%hu): %u",
			RTA_ALIGN(ifla[IFLA_NUM_VF]->rta_len),
			*(unsigned *)RTA_DATA(ifla[IFLA_NUM_VF]));

	/*
	 * debug VFs information list
	 *
	 * IFLA_VFINFO_LIST was committed in kernel 2.6.34.
	 *
	 * commit c02db8c6290bb992442fec1407643c94cc414375
	 * Author: Chris Wright <chrisw@sous-sol.org>
	 * Date:   Sun May 16 01:05:45 2010 -0700
	 *
	 *     rtnetlink: make SR-IOV VF interface symmetric
	 */ 
	if(ifla[IFLA_VFINFO_LIST])
		rec_dbg("    IFLA_VFINFO_LIST(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_VFINFO_LIST]->rta_len));

	/*
	 * debug 64-bit interface statistics counter
	 *
	 * IFLA_STATS64 was committed in kernel 2.6.35.
	 *
	 * commit 10708f37ae729baba9b67bd134c3720709d4ae62
	 * Author: Jan Engelhardt <jengelh@medozas.de>
	 * Date:   Thu Mar 11 09:57:29 2010 +0000
	 *
	 *     net: core: add IFLA_STATS64 support
	 */
	if(ifla[IFLA_STATS64])
		rec_dbg("    IFLA_STATS64(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_STATS64]->rta_len));

	/*
	 * debug virtual port management
	 *
	 * IFLA_VF_PORTS and IFLA_PORT_SELF was committed in kernel 2.6.35.
	 *
	 * commit 57b610805ce92dbd79fc97509f80fa5391b99623
	 * Author: Scott Feldman <scofeldm@cisco.com>
	 * Date:   Mon May 17 22:49:55 2010 -0700
	 *
	 *     net: Add netlink support for virtual port management (was iovnl)
	 */
	if(ifla[IFLA_VF_PORTS])
		rec_dbg("    IFLA_VF_PORTS(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_VF_PORTS]->rta_len));

	if(ifla[IFLA_PORT_SELF])
		rec_dbg("    IFLA_PORT_SELF(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_PORT_SELF]->rta_len));

	/*
	 * debug link address family
	 *
	 * IFLA_AF_SPEC was committed in kernel 2.6.38.
	 *
	 * commit f8ff182c716c6f11ca3061961f5722f26a14e101
	 * Author: Thomas Graf <tgraf@infradead.org>
	 * Date:   Tue Nov 16 04:30:14 2010 +0000
	 *
	 *     rtnetlink: Link address family API
	 */
	if(ifla[IFLA_AF_SPEC])
		rec_dbg("    IFLA_AF_SPEC(%hu): -- ignored --",
			RTA_ALIGN(ifla[IFLA_AF_SPEC]->rta_len));

	rec_dbg("");

	return;
}

/*
 * debug VLAN interface information messages
 */
void debug_vlaninfo(struct rtattr *vlaninfo[])
{
	/* debug VLAN ID */
	if(vlaninfo[IFLA_VLAN_ID])
		rec_dbg("            IFLA_VLAN_ID(%hu): %d",
			RTA_ALIGN(vlaninfo[IFLA_VLAN_ID]->rta_len),
			*(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]));
	
	/* debug VLAN flags */
	if(vlaninfo[IFLA_VLAN_FLAGS]) {
		struct ifla_vlan_flags *vlan_flags =
			(struct ifla_vlan_flags *)RTA_DATA(vlaninfo[IFLA_VLAN_FLAGS]);
		char flags_list[MAX_STR_SIZE] = "";
		rec_dbg("            IFLA_VLAN_FLAGS(%hu):",
			RTA_ALIGN(vlaninfo[IFLA_VLAN_FLAGS]->rta_len));
		rec_dbg("                [ ifla_vlan_flags(%d) ]",
			NLMSG_ALIGN(sizeof(struct ifla_vlan_flags)));
		debug_n2s_vlan_flags(vlan_flags->flags, flags_list);
		rec_dbg("                    flags(%d): 0x%.8x(%s)",
			sizeof(vlan_flags->flags), vlan_flags->flags, flags_list);
		rec_dbg("                    mask(%d): 0x%.8x",
			sizeof(vlan_flags->mask), vlan_flags->mask);
	}

	/* debug VLAN egress QoS */
	if(vlaninfo[IFLA_VLAN_EGRESS_QOS]) {
		struct rtattr *vlan_egress_qos[__IFLA_VLAN_QOS_MAX];
		rec_dbg("            IFLA_VLAN_EGRESS_QOS(%hu):",
			RTA_ALIGN(vlaninfo[IFLA_VLAN_EGRESS_QOS]->rta_len));
		parse_vlanqos(vlan_egress_qos, vlaninfo[IFLA_VLAN_EGRESS_QOS]);
		debug_vlanqos(vlan_egress_qos);
	}

	/* debug VLAN ingress QoS */
	if(vlaninfo[IFLA_VLAN_INGRESS_QOS]) {
		struct rtattr *vlan_ingress_qos[__IFLA_VLAN_QOS_MAX];
		rec_dbg("            IFLA_VLAN_INGRESS_QOS(%hu):",
			RTA_ALIGN(vlaninfo[IFLA_VLAN_INGRESS_QOS]->rta_len));
		parse_vlanqos(vlan_ingress_qos, vlaninfo[IFLA_VLAN_INGRESS_QOS]);
		debug_vlanqos(vlan_ingress_qos);
	}
}

/*
 * debug VLAN QoS informaton messages
 */
void debug_vlanqos(struct rtattr *vlanqos[])
{
	if(vlanqos[IFLA_VLAN_QOS_MAPPING]) {
		struct ifla_vlan_qos_mapping *vlan_qos_map =
			(struct ifla_vlan_qos_mapping *)RTA_DATA(vlanqos[IFLA_VLAN_QOS_MAPPING]);
		rec_dbg("                IFLA_VLAN_QOS_MAPPING(%hu):",
			RTA_ALIGN(vlanqos[IFLA_VLAN_QOS_MAPPING]->rta_len));
		rec_dbg("                    [ ifla_vlan_qos_mapping(%d) ]",
			sizeof(struct ifla_vlan_qos_mapping));
		rec_dbg("                        from(%d): %u",
			sizeof(vlan_qos_map->from), vlan_qos_map->from);
		rec_dbg("                        to(%d): %u",
			sizeof(vlan_qos_map->to), vlan_qos_map->to);
	}
}

/*
 * debug GRE interface information messages
 */
void debug_greinfo(struct rtattr *greinfo[])
{
	/* debug link in GRE tunnel */
	if(greinfo[IFLA_GRE_LINK]) {
		unsigned char gre_ifindex =
			*(unsigned char *)RTA_DATA(greinfo[IFLA_GRE_LINK]);
		char gre_ifname[IFNAMSIZ] = "";

		if(!if_indextoname(gre_ifindex, gre_ifname))
			if(!if_indextoname_from_iflist(gre_ifindex, gre_ifname))
				if_indextoname_from_ifhist(gre_ifindex, gre_ifname);
		rec_dbg("            IFLA_GRE_LINK(%hu): %u(%s)",
			RTA_ALIGN(greinfo[IFLA_GRE_LINK]->rta_len),
			gre_ifindex, gre_ifname);
	}

	/* debug iflags in GRE tunnel */
	if(greinfo[IFLA_GRE_IFLAGS]) {
		char iflags[32] = "";

		debug_n2s_gre_flags(*(unsigned short *)RTA_DATA(greinfo[IFLA_GRE_IFLAGS]), iflags);
		rec_dbg("            IFLA_GRE_IFLAGS(%hu): 0x%.4x(%s)",
			RTA_ALIGN(greinfo[IFLA_GRE_IFLAGS]->rta_len),
			*(unsigned short *)RTA_DATA(greinfo[IFLA_GRE_IFLAGS]),
			iflags);
	}

	/* debug oflags in GRE tunnel */
	if(greinfo[IFLA_GRE_OFLAGS]) {
		char oflags[32] = "";

		debug_n2s_gre_flags(*(unsigned short *)RTA_DATA(greinfo[IFLA_GRE_OFLAGS]), oflags);
		rec_dbg("            IFLA_GRE_OFLAGS(%hu): 0x%.4x(%s)",
			RTA_ALIGN(greinfo[IFLA_GRE_OFLAGS]->rta_len),
			*(unsigned short *)RTA_DATA(greinfo[IFLA_GRE_OFLAGS]),
			oflags);
	}

	/* debug ikey in GRE tunnel */
	if(greinfo[IFLA_GRE_IKEY])
		rec_dbg("            IFLA_GRE_IKEY(%hu): 0x%.8x",
			RTA_ALIGN(greinfo[IFLA_GRE_IKEY]->rta_len),
			*(unsigned *)RTA_DATA(greinfo[IFLA_GRE_IKEY]));

	/* debug okey in GRE tunnel */
	if(greinfo[IFLA_GRE_OKEY])
		rec_dbg("            IFLA_GRE_OKEY(%hu): 0x%.8x",
			RTA_ALIGN(greinfo[IFLA_GRE_OKEY]->rta_len),
			*(unsigned *)RTA_DATA(greinfo[IFLA_GRE_OKEY]));

	/* debug local address in GRE tunnel */
	if(greinfo[IFLA_GRE_LOCAL]) {
		char gre_laddr[INET6_ADDRSTRLEN+1] = "";

		inet_ntop(AF_INET, RTA_DATA(greinfo[IFLA_GRE_LOCAL]),
			gre_laddr, sizeof(gre_laddr));
		rec_dbg("            IFLA_GRE_LOCAL(%hu): %s",
			RTA_ALIGN(greinfo[IFLA_GRE_LOCAL]->rta_len), gre_laddr);
	}

	/* debug remote address in GRE tunnel */
	if(greinfo[IFLA_GRE_REMOTE]) {
		char gre_raddr[INET6_ADDRSTRLEN+1] = "";

		inet_ntop(AF_INET, RTA_DATA(greinfo[IFLA_GRE_REMOTE]),
			gre_raddr, sizeof(gre_raddr));
		rec_dbg("            IFLA_GRE_REMOTE(%hu): %s",
			RTA_ALIGN(greinfo[IFLA_GRE_REMOTE]->rta_len), gre_raddr);
	}

	/* debug TTL in GRE tunnel */
	if(greinfo[IFLA_GRE_TTL])
		rec_dbg("            IFLA_GRE_TTL(%hu): %u",
			RTA_ALIGN(greinfo[IFLA_GRE_TTL]->rta_len),
			*(unsigned char *)RTA_DATA(greinfo[IFLA_GRE_TTL]));

	/* debug TOS in GRE tunnel */
	if(greinfo[IFLA_GRE_TOS])
		rec_dbg("            IFLA_GRE_TOS(%hu): %u",
			RTA_ALIGN(greinfo[IFLA_GRE_TOS]->rta_len),
			*(unsigned char *)RTA_DATA(greinfo[IFLA_GRE_TOS]));

	/* debug path MTU discovery in GRE tunnel */
	if(greinfo[IFLA_GRE_PMTUDISC])
		rec_dbg("            IFLA_GRE_PMTUDISC(%hu): %u",
			RTA_ALIGN(greinfo[IFLA_GRE_PMTUDISC]->rta_len),
			*(unsigned char *)RTA_DATA(greinfo[IFLA_GRE_PMTUDISC]));
}

/*
 * convert interafce type from number to string
 */ 
const char *debug_n2s_ifi_type(int type)
{
#define _IFI_TYPE(s) \
	if(type == ARPHRD_##s) \
		return #s;
	_IFI_TYPE(NETROM);
	_IFI_TYPE(ETHER);
	_IFI_TYPE(EETHER);
	_IFI_TYPE(AX25);
	_IFI_TYPE(PRONET);
	_IFI_TYPE(CHAOS);
	_IFI_TYPE(IEEE802);
	_IFI_TYPE(ARCNET);
	_IFI_TYPE(APPLETLK);
	_IFI_TYPE(DLCI);
	_IFI_TYPE(ATM);
	_IFI_TYPE(METRICOM);
	_IFI_TYPE(IEEE1394);
	_IFI_TYPE(EUI64);
	_IFI_TYPE(INFINIBAND);
	_IFI_TYPE(SLIP);
	_IFI_TYPE(CSLIP);
	_IFI_TYPE(SLIP6);
	_IFI_TYPE(CSLIP6);
	_IFI_TYPE(RSRVD);
	_IFI_TYPE(ADAPT);
	_IFI_TYPE(ROSE);
	_IFI_TYPE(X25);
	_IFI_TYPE(HWX25);
	_IFI_TYPE(CAN);
	_IFI_TYPE(PPP);
	_IFI_TYPE(CISCO);
	_IFI_TYPE(HDLC);
	_IFI_TYPE(LAPB);
	_IFI_TYPE(DDCMP);
	_IFI_TYPE(RAWHDLC);
	_IFI_TYPE(TUNNEL);
	_IFI_TYPE(TUNNEL6);
	_IFI_TYPE(FRAD);
	_IFI_TYPE(SKIP);
	_IFI_TYPE(LOOPBACK);
	_IFI_TYPE(LOCALTLK);
	_IFI_TYPE(FDDI);
	_IFI_TYPE(BIF);
	_IFI_TYPE(SIT);
	_IFI_TYPE(IPDDP);
	_IFI_TYPE(IPGRE);
	_IFI_TYPE(PIMREG);
	_IFI_TYPE(HIPPI);
	_IFI_TYPE(ASH);
	_IFI_TYPE(ECONET);
	_IFI_TYPE(IRDA);
	_IFI_TYPE(FCPP);
	_IFI_TYPE(FCAL);
	_IFI_TYPE(FCPL);
	_IFI_TYPE(FCFABRIC);
	_IFI_TYPE(IEEE802_TR);
	_IFI_TYPE(IEEE80211);
	_IFI_TYPE(IEEE80211_PRISM);
	_IFI_TYPE(IEEE80211_RADIOTAP);
	_IFI_TYPE(IEEE802154);
	_IFI_TYPE(PHONET);
	_IFI_TYPE(PHONET_PIPE);
#undef _IFI_TYPE
}

/*
 * convert interface flags from number to string
 */
void debug_n2s_ifi_flags(int flags, char *flags_list)
{
#define _IFI_FLAGS(s) \
	if(flags & IFF_##s) \
		(flags &= ~IFF_##s) ? strcat(flags_list, #s ",") : strcat(flags_list, #s);
	_IFI_FLAGS(UP);
	_IFI_FLAGS(BROADCAST);
	_IFI_FLAGS(DEBUG);
	_IFI_FLAGS(LOOPBACK);
	_IFI_FLAGS(POINTOPOINT);
	_IFI_FLAGS(NOTRAILERS);
	_IFI_FLAGS(RUNNING);
	_IFI_FLAGS(NOARP);
	_IFI_FLAGS(PROMISC);
	_IFI_FLAGS(ALLMULTI);
	_IFI_FLAGS(MASTER);
	_IFI_FLAGS(SLAVE);
	_IFI_FLAGS(MULTICAST);
	_IFI_FLAGS(PORTSEL);
	_IFI_FLAGS(AUTOMEDIA);
	_IFI_FLAGS(DYNAMIC);
	_IFI_FLAGS(LOWER_UP);
	_IFI_FLAGS(DORMANT);
	_IFI_FLAGS(ECHO);
#undef _IFI_FLAGS
}

/*
 * convert interafce operational state from number to string
 */
const char *debug_n2s_if_oper_state(int state)
{
#define _IF_OPER_STATE(s) \
	if(state == IF_OPER_##s) \
		return #s;
	_IF_OPER_STATE(UNKNOWN);
	_IF_OPER_STATE(NOTPRESENT);
	_IF_OPER_STATE(DOWN);
	_IF_OPER_STATE(LOWERLAYERDOWN);
	_IF_OPER_STATE(TESTING);
	_IF_OPER_STATE(DORMANT);
	_IF_OPER_STATE(UP);
#undef _IF_OPER_STATE
}

/*
 * convert interafce link mode from number to string
 */
const char *debug_n2s_ifla_linkmode(int mode)
{
#define _IFLA_LINKMODE(s) \
	if(mode == IF_LINK_MODE_##s) \
		return #s;
	_IFLA_LINKMODE(DEFAULT);
	_IFLA_LINKMODE(DORMANT);
#undef _IFLA_LINKMODE
}

/*
 * convert VLAN flags from number to string
 */
void debug_n2s_vlan_flags(int flags, char *flags_list)
{
#define _VLAN_FLAGS(s) \
	if(flags & VLAN_FLAG_##s) \
		(flags &= ~VLAN_FLAG_##s) ? strcat(flags_list, #s ",") : strcat(flags_list, #s);
	_VLAN_FLAGS(REORDER_HDR);
	_VLAN_FLAGS(GVRP);
	_VLAN_FLAGS(LOOSE_BINDING);
#undef _VLAN_FLAGS
}

/*
 * convert GRE flags from number to string
 */
void debug_n2s_gre_flags(int flags, char *flags_list)
{
#define _GRE_FLAGS(s) \
	if(flags & GRE_##s) \
		(flags &= ~GRE_##s) ? strcat(flags_list, #s ",") : strcat(flags_list, #s);
	_GRE_FLAGS(CSUM);
	_GRE_FLAGS(ROUTING);
	_GRE_FLAGS(KEY);
	_GRE_FLAGS(SEQ);
	_GRE_FLAGS(STRICT);
#undef _GRE_FLAGS
}
