#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/* top of interface list */
static LIST_HEAD(head);

/*
 * create a neighbor discovery cache list
 */ 
int create_ndlist(struct msghdr *msg)
{
	struct nlmsghdr *nlh;
	int nlh_len;
	struct ndmsg *ndm;
	int ndm_len;
	struct rtattr *nda[__NDA_MAX];
	struct ndlist_entry *ndle;

	/* get netlink message header */
	nlh = msg->msg_iov->iov_base;
	nlh_len = msg->msg_iov->iov_len;

	/* parse netlink message header */
	for( ; NLMSG_OK(nlh, nlh_len); nlh = NLMSG_NEXT(nlh, nlh_len)) {
		/* whether netlink message header ends or not */
		if(nlh->nlmsg_type == NLMSG_DONE)
			return(1);
	
		/* get neighbor discovery message */
		ndm = (struct ndmsg *)NLMSG_DATA(nlh);
		ndm_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);

		/* parse neighbor discovery attributes */
		parse_ndisc(nda, nlh);

		/* debug */
		if(optflag & F_DEBUG) {
			debug_nlmsg(nlh);
			debug_ndmsg(ndm, nda, ndm_len);
		}

		/* check address family */
		if((ndm->ndm_family != AF_INET) && (ndm->ndm_family != AF_INET6))
			continue;

		/* check address type */
		if(ndm->ndm_type != RTN_UNICAST)
			continue;

		/* create neighbor discovery list */
		ndle = (struct ndlist_entry *)malloc(sizeof(struct ndlist_entry));
		if(ndle == NULL) {
			rec_log("[%-12s] %s: malloc: failed", "ERROR", __func__);
			return(-1);
		}
		memset(ndle, 0, sizeof(struct ndlist_entry));
		INIT_LIST_HEAD(&(ndle->list));

		/* get interface index */
		ndle->ifindex = ndm->ndm_ifindex;

		/* convert interface index to name */
		if_indextoname(ndm->ndm_ifindex, ndle->ifname);

		/* get ip address */
		if(nda[NDA_DST])
			inet_ntop(ndm->ndm_family, (struct in_addr *)RTA_DATA(nda[NDA_DST]),
				ndle->ipaddr, sizeof(ndle->ipaddr));
	
		/* get link local address */
		if(nda[NDA_LLADDR])
			memcpy(ndle->lladdr, RTA_DATA(nda[NDA_LLADDR]), sizeof(ndle->lladdr));

		/* add neighbor discovery cache list */
		list_add(&(ndle->list), &head);
	}

	//print_ndlist();

	return(0);
}

/*
 * print a neighbor discovery cache list
 */ 
void print_ndlist(void)
{
	struct list_head *l;

	list_for_each(l, &head) {
		struct ndlist_entry *e;

		e = list_entry(l, struct ndlist_entry, list);

		printf("ifindex: %d, ", e->ifindex);
		printf("ifname: %s, ", e->ifname);
		printf("ipaddr: %s, ", e->ipaddr);
		printf("lladdr: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
			e->lladdr[0], e->lladdr[1], e->lladdr[2], e->lladdr[3], e->lladdr[4], e->lladdr[5]);
	}
}

/*
 * search an entry in a neighbor discovery cache list
 */ 
static inline struct ndlist_entry *search_ndlist(char *ipa)
{
	struct list_head *l;

	list_for_each(l, &head) {
		struct ndlist_entry *e;

		e = list_entry(l, struct ndlist_entry, list);
		if(!memcmp(ipa, &(e->ipaddr), sizeof(e->ipaddr)))
			return e;
	}

	return(NULL);
}

/*
 * add a neighbor discovery cache to a list
 */
static inline void add_ndlist_entry(struct list_head *l)
{
	INIT_LIST_HEAD(l);
	list_add(l, &head);
}

/*
 * [not in use]
 * delete an entry in a interface list entry
 */
static inline void del_iflist_entry(char *ipa)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &head) {
		struct ndlist_entry *e;

		e = list_entry(l, struct ndlist_entry, list);
		if(!memcmp(ipa, &(e->ipaddr), sizeof(e->ipaddr))) {
			list_del(l);
			free(e);
		}
	}
}

/*
 * parse neighbor discovery messages
 */
int parse_ndmsg(struct nlmsghdr *nlh)
{
	struct ndmsg *ndm = (struct ndmsg *)NLMSG_DATA(nlh);
	int ndm_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);
	struct rtattr *nda[__NDA_MAX];
	struct ndlist_entry *ndle_tmp, *ndle;
	char msg_type[MAX_STR_SIZE] = "";
	char ndm_type[MAX_STR_SIZE] = "";

	/* parse neighbor discovery attributes */
	parse_ndisc(nda, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ndmsg(ndm, nda, ndm_len);
	}

	/* check address family */
	if(ndm->ndm_family == AF_INET) {
		if((optflag & F_IPV4) || !(optflag & F_IPV6)) {
			strcpy(msg_type, "IPv4 ARP");
			strcpy(ndm_type, "arp cache");
		} else {
			return(1);
		}
	} else if(ndm->ndm_family == AF_INET6) {
		if((optflag & F_IPV6) || !(optflag & F_IPV4)) {
			strcpy(msg_type, "IPv6 NDP");
			strcpy(ndm_type, "ndp cache");
		} else {
			return(1);
		}
	} else {
		return(1);
	}

	/* unicast only */
	if(ndm->ndm_type != RTN_UNICAST)
		return(1);

	/* create temporary neighbor discovery entry */
	ndle_tmp = malloc(sizeof(struct ndlist_entry));
	if(!ndle_tmp) {
		rec_log("[%-12s] %s: malloc() failed", "ERROR", __func__);
		return(1);
	}
	memset(ndle_tmp, 0, sizeof(struct ndlist_entry));

	/* get interface index */
	ndle_tmp->ifindex = ndm->ndm_ifindex;

	/* convert interface index to name */
	if_indextoname((unsigned)ndm->ndm_ifindex, ndle_tmp->ifname);

	/* get ip address */
	if(nda[NDA_DST])
		inet_ntop(ndm->ndm_family, (struct in_addr *)RTA_DATA(nda[NDA_DST]),
			ndle_tmp->ipaddr, sizeof(ndle_tmp->ipaddr));

	/* get link local address */
	if(nda[NDA_LLADDR])
		memcpy(ndle_tmp->lladdr, RTA_DATA(nda[NDA_LLADDR]), sizeof(ndle_tmp->lladdr));

	/* logging neighbor discovery message */
	if(nlh->nlmsg_type == RTM_NEWNEIGH) {
		/* search neighbor discovery entry by ip address */
		ndle = search_ndlist(ndle_tmp->ipaddr);
		switch(ndm->ndm_state) {
			case NUD_FAILED:
				/* change or add neighbor discovery list entry */
				if(ndle) {
					if(memcmp(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr))) {
						memcpy(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr));
						rec_log("[%-12s] %s invalidated: if=%s ipaddr=%s "
							"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
							msg_type, ndm_type, ndle->ifname, ndle->ipaddr,
							ndle->lladdr[0], ndle->lladdr[1], ndle->lladdr[2],
							ndle->lladdr[3], ndle->lladdr[4], ndle->lladdr[5]);
					}
					free(ndle_tmp);
				} else {
					add_ndlist_entry(&(ndle_tmp->list));
					rec_log("[%-12s] %s invalidated: if=%s ipaddr=%s "
						"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
						msg_type, ndm_type, ndle_tmp->ifname, ndle_tmp->ipaddr,
						ndle_tmp->lladdr[0], ndle_tmp->lladdr[1], ndle_tmp->lladdr[2],
						ndle_tmp->lladdr[3], ndle_tmp->lladdr[4], ndle_tmp->lladdr[5]);
				}
				break;
			case NUD_PERMANENT:
			case NUD_PROBE:
			case NUD_DELAY:
			case NUD_STALE:
			case NUD_REACHABLE:
				/* change or add neighbor discovery list entry */
				if(ndle) {
					if(memcmp(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr))) {
						memcpy(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr));
						rec_log("[%-12s] %s changed: if=%s ipaddr=%s "
							"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
							msg_type, ndm_type, ndle->ifname, ndle->ipaddr,
							ndle->lladdr[0], ndle->lladdr[1], ndle->lladdr[2],
							ndle->lladdr[3], ndle->lladdr[4], ndle->lladdr[5]);
					}
					free(ndle_tmp);
				} else {
					add_ndlist_entry(&(ndle_tmp->list));
					rec_log("[%-12s] %s added: if=%s ipaddr=%s "
						"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
						msg_type, ndm_type, ndle_tmp->ifname, ndle_tmp->ipaddr,
						ndle_tmp->lladdr[0], ndle_tmp->lladdr[1], ndle_tmp->lladdr[2],
						ndle_tmp->lladdr[3], ndle_tmp->lladdr[4], ndle_tmp->lladdr[5]);
				}
				break;
			case NUD_NOARP:
			case NUD_INCOMPLETE:
				break;
		}
	} else if(nlh->nlmsg_type == RTM_DELNEIGH) {
		struct list_head *l, *n;
		struct ndlist_entry *e;

		switch(ndm->ndm_state) {
			case NUD_FAILED:
			case NUD_INCOMPLETE:
			case NUD_PERMANENT:
			case NUD_REACHABLE:
			case NUD_STALE:
			case NUD_DELAY:
			case NUD_PROBE:
			case NUD_NOARP:
				/* search & delete neighbor discovery entry */
				list_for_each_safe(l, n, &head) {
					e = list_entry(l, struct ndlist_entry, list);
					if(!memcmp(e->ipaddr, ndle_tmp->ipaddr, sizeof(e->ipaddr))) {
						rec_log("[%-12s] %s deleted: if=%s ipaddr=%s "
							"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
							msg_type, ndm_type, e->ifname, ndle_tmp->ipaddr,
							ndle_tmp->lladdr[0], ndle_tmp->lladdr[1],
							ndle_tmp->lladdr[2], ndle_tmp->lladdr[3],
							ndle_tmp->lladdr[4], ndle_tmp->lladdr[5]);
						list_del(l);
						free(e);
						/* because entry matched is only one */
						break;
					}
				}
			break;
		}
		free(ndle_tmp);
	}

	return(0);
}

/*
 * debug neighbor discovery messages
 */
void debug_ndmsg(struct ndmsg *ndm, struct rtattr *nda[], int ndm_len)
{
	/* debug ndmsg */
	char ifname[IFNAMSIZ] = "";
	char ndm_flags_list[MAX_STR_SIZE] = "";

	rec_dbg("*********************************************************************");
	rec_dbg("[ ndmsg(%d) ]",
		NLMSG_ALIGN(sizeof(struct ndmsg)));
	rec_dbg("    ndm_family(%d): %d(%s)",
		sizeof(ndm->ndm_family), ndm->ndm_family, debug_n2s_af_type(ndm->ndm_family));
	if_indextoname((unsigned)ndm->ndm_ifindex, ifname);
	rec_dbg("    ndm_ifindex(%d): %d(%s)",
		sizeof(ndm->ndm_ifindex), ndm->ndm_ifindex, ifname);
	rec_dbg("    ndm_state(%d): %d(%s)",
		sizeof(ndm->ndm_state), ndm->ndm_state, debug_n2s_ndm_state(ndm->ndm_state));
	rec_dbg("    ndm_flags(%d): %d(%s)",
		sizeof(ndm->ndm_flags), ndm->ndm_flags, ndm_flags_list);
	debug_n2s_ndm_flags(ndm->ndm_flags, ndm_flags_list);
	rec_dbg("    ndm_type(%d): %d(%s)",
		sizeof(ndm->ndm_type), ndm->ndm_type, debug_n2s_ndm_type(ndm->ndm_type));

	/* debug neighbor discovery attributes */
	rec_dbg("*********************************************************************");
	rec_dbg("[ ndmsg attributes(%d) ]",
		NLMSG_ALIGN(ndm_len - NLMSG_ALIGN(sizeof(struct ndmsg))));

	/* debug ip address */
	if(nda[NDA_DST]) {
		char ipa[INET6_ADDRSTRLEN] = "";
		inet_ntop(ndm->ndm_family, (struct in_addr *)RTA_DATA(nda[NDA_DST]), ipa, sizeof(ipa));
		rec_dbg("    NDA_DST(%hu): %s",
			RTA_ALIGN(nda[NDA_DST]->rta_len), ipa);
	}

	/* debug link local address */
	if(nda[NDA_LLADDR]) {
		unsigned char lla[ETHER_ADDR_LEN] = "";
		memcpy(lla, RTA_DATA(nda[NDA_LLADDR]), ETHER_ADDR_LEN);
		rec_dbg("    NDA_LLADDR(%hu): %.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
			RTA_ALIGN(nda[NDA_LLADDR]->rta_len),
			lla[0], lla[1], lla[2], lla[3], lla[4], lla[5]);
	}

	/* debug neighbor discovery cache information */
	if(nda[NDA_CACHEINFO]) {
		struct nda_cacheinfo *ndac;
		ndac = (struct nda_cacheinfo *)RTA_DATA(nda[NDA_CACHEINFO]);
		rec_dbg("    NDA_CACHEINFO(%hu):",
			RTA_ALIGN(nda[NDA_CACHEINFO]->rta_len));
		rec_dbg("        [ nda_cacheinfo(%d) ]",
			NLMSG_ALIGN(sizeof(struct nda_cacheinfo)));
		rec_dbg("            ndm_confirmed(%d): %u",
			sizeof(ndac->ndm_confirmed), ndac->ndm_confirmed);
		rec_dbg("            ndm_used(%d): %u",
			sizeof(ndac->ndm_used), ndac->ndm_used);
		rec_dbg("            ndm_updated(%d): %u",
			sizeof(ndac->ndm_updated), ndac->ndm_updated);
		rec_dbg("            ndm_refcnt(%d): %u",
			sizeof(ndac->ndm_refcnt), ndac->ndm_refcnt);
	}

	/* debug neighbor discovery probes */
	if(nda[NDA_PROBES])
		rec_dbg("    NDA_PROBES(%hu): %u",
			RTA_ALIGN(nda[NDA_PROBES]->rta_len), *(unsigned *)RTA_DATA(nda[NDA_PROBES]));

	rec_dbg("");

	return;
}

const char *debug_n2s_ndm_state(int state)
{
#define _NDM_STATE(s) \
	if(state == NUD_##s) \
		return #s;
	_NDM_STATE(INCOMPLETE);
	_NDM_STATE(REACHABLE);
	_NDM_STATE(STALE);
	_NDM_STATE(DELAY);
	_NDM_STATE(PROBE);
	_NDM_STATE(FAILED);
	_NDM_STATE(NOARP);
	_NDM_STATE(PERMANENT);
#undef _NDM_STATE
}

void debug_n2s_ndm_flags(int flags, char *flags_list)
{
#define _NDM_FLAGS(f) \
	if(flags & NTF_##f) \
		(flags &= ~NTF_##f) ? strcat(flags_list, #f ",") : strcat(flags_list, #f);
	_NDM_FLAGS(USE);
	_NDM_FLAGS(PROXY);
	_NDM_FLAGS(ROUTER);
#undef _NDM_FLAGS
}

const char *debug_n2s_ndm_type(int type) 
{
#define _NDM_TYPE(s) \
	if(type == RTN_##s) \
		return #s;
	_NDM_TYPE(UNICAST);
	_NDM_TYPE(LOCAL);
	_NDM_TYPE(BROADCAST);
	_NDM_TYPE(ANYCAST);
	_NDM_TYPE(MULTICAST);
	_NDM_TYPE(BLACKHOLE);
	_NDM_TYPE(UNREACHABLE);
	_NDM_TYPE(PROHIBIT);
	_NDM_TYPE(THROW);
	_NDM_TYPE(NAT);
	_NDM_TYPE(XRESOLVE);
#undef _NDM_TYPE
}
