#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/*
 * parse route messages
 */
int parse_rtmsg(struct nlmsghdr *nlh)
{
	struct rtmsg *rtm = (struct rtmsg *)NLMSG_DATA(nlh);
	int rtm_len = NLMSG_ALIGN(nlh->nlmsg_len - NLMSG_HDRLEN);
	struct rtattr *rta[__RTA_MAX];
	char msg_type[MAX_STR_SIZE] = "", msg[MAX_MSG_SIZE] = "";
	char *mp = msg;

	/* parse route attributes */
	parse_rtattr(rta, RTA_MAX, RTM_RTA(rtm), RTM_PAYLOAD(nlh));

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_rtmsg(rtm, rta, rtm_len);
	}

	/* check address family */
	char dst[INET6_ADDRSTRLEN] = "";
	if(rtm->rtm_family == AF_INET) {
		strcpy(msg_type, "IPv4 ROUTE");
		strcpy(dst, "0.0.0.0");
	} else if(rtm->rtm_family == AF_INET6) {
		strcpy(msg_type, "IPv6 ROUTE");
		strcpy(dst, "::");
	} else {
		rec_log("[%-12s] %s: unknown address family: %d",
			"ERROR", __func__, rtm->rtm_family);
		return(1);
	}

	/* convert from table id to table name */
	char table[MAX_STR_SIZE] = "";
	parse_n2s_rtm_table(rtm->rtm_table, table);

	/* check route table id(other than RT_TABLE_LOCAL) */
	if(rtm->rtm_table == RT_TABLE_LOCAL)
		return(1);

	/* check route protocol(other than RTPROT_UNSPEC) */
	if(rtm->rtm_protocol == RTPROT_UNSPEC)
		return(1);
	
	/* check route flags(other then RTM_F_CLONED */
	if(rtm->rtm_flags & RTM_F_CLONED)
		return(1);

	/* get destination prefix */
	if(rta[RTA_DST])
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_DST]), dst, sizeof(dst));
	mp += sprintf(mp, "dst=%s/%d ", dst, rtm->rtm_dst_len);

	/* get source prefix */
	if(rta[RTA_SRC]) {
		char src[INET6_ADDRSTRLEN] = "";

		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_SRC]), src, sizeof(src));
		mp += sprintf(mp, "src=%s/%d ", src, rtm->rtm_src_len);
	}

	/* get preferred source address */
	if(rta[RTA_PREFSRC]) {
		char prefsrc[INET6_ADDRSTRLEN] = "";

		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_PREFSRC]), prefsrc, sizeof(prefsrc));
		mp += sprintf(mp, "prefsrc=%s ", prefsrc);
	}

	/* get tos */
	if(rtm->rtm_tos)
		mp += sprintf(mp, "tos=0x%.2x ", rtm->rtm_tos);

	/* get ingress interface */
	if(rta[RTA_IIF]) {
		unsigned iifindex;
		char iifname[IFNAMSIZ] = "";

		iifindex = *((unsigned *)RTA_DATA(rta[RTA_IIF]));
		if(nlh->nlmsg_type == RTM_NEWROUTE)
			if_indextoname(iifindex, iifname);
		else if(nlh->nlmsg_type == RTM_DELROUTE)
			if(!if_indextoname_from_iflist(iifindex, iifname))
				if_indextoname_from_ifhist(iifindex, iifname);
		mp += sprintf(mp, "in=%s ", iifname);
	}

	/* get gateway address */
	if(rta[RTA_GATEWAY]) {
		char gw[INET6_ADDRSTRLEN] = "";

		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_GATEWAY]), gw, sizeof(gw));
		mp += sprintf(mp, "gw=%s ", gw);
	}

	/* get egress interface */
	if(rta[RTA_OIF]) {
		unsigned oifindex;
		char oifname[IFNAMSIZ] = "";

		oifindex = *((unsigned *)RTA_DATA(rta[RTA_OIF]));
		if(nlh->nlmsg_type == RTM_NEWROUTE)
			if_indextoname(oifindex, oifname);
		else if(nlh->nlmsg_type == RTM_DELROUTE)
			if(!if_indextoname_from_iflist(oifindex, oifname))
				if_indextoname_from_ifhist(oifindex, oifname);
		mp += sprintf(mp, "out=%s ", oifname);
	}

	/* get priority(but metric) */
	char metric[MAX_STR_SIZE] = "";
	if(rta[RTA_PRIORITY])
		sprintf(metric, "metric=%d ", *((int *)RTA_DATA(rta[RTA_PRIORITY])));

	/* convert route message type */
	char type[MAX_STR_SIZE] = "";
	parse_n2s_rtm_type(rtm->rtm_type, type);

	/* convert route message protocol */
	char proto[MAX_STR_SIZE] = "";
	parse_n2s_rtm_protocol(rtm->rtm_protocol, proto);

	/* get table id & name */
	if(rta[RTA_TABLE])
		parse_n2s_rtm_table(*(int *)RTA_DATA(rta[RTA_TABLE]), table);

	/* get multipath */
	if(rta[RTA_MULTIPATH]) {
		struct rtnexthop *rtnh = RTA_DATA(rta[RTA_MULTIPATH]);
		int rtnh_len = RTA_PAYLOAD(rta[RTA_MULTIPATH]);
		struct rtattr *rtna[__RTA_MAX];
		char rtnh_ifname[IFNAMSIZ] = "";
		char rtnh_gw[INET6_ADDRSTRLEN] = "";

		for(; RTNH_OK(rtnh, rtnh_len);
			rtnh = RTNH_NEXT(rtnh), rtnh_len -= RTNH_ALIGN(rtnh->rtnh_len)) {
			parse_rtattr(rtna, RTA_MAX, RTNH_DATA(rtnh), rtnh->rtnh_len - sizeof(*rtnh));
			if(rtna[RTA_GATEWAY])
				inet_ntop(rtm->rtm_family, RTA_DATA(rtna[RTA_GATEWAY]),
					rtnh_gw, sizeof(rtnh_gw));

			/* get interface name & logging routing table message */
			if(nlh->nlmsg_type == RTM_NEWROUTE) {
				if_indextoname(rtnh->rtnh_ifindex, rtnh_ifname);

				rec_log("[%-12s] route added: %sgw=%s out=%s "
					"%sweight=%d type=%s proto=%s table=%s",
					msg_type, msg, rtnh_gw, rtnh_ifname,
					metric, rtnh->rtnh_hops+1, type, proto, table);
			} else if(nlh->nlmsg_type == RTM_DELROUTE) {
				if(!if_indextoname_from_iflist(rtnh->rtnh_ifindex, rtnh_ifname))
					if_indextoname_from_ifhist(rtnh->rtnh_ifindex, rtnh_ifname);

				rec_log("[%-12s] route deleted: %sgw=%s out=%s " 
					"%sweight=%d type=%s proto=%s table=%s",
					msg_type, msg, rtnh_gw, rtnh_ifname,
					metric, rtnh->rtnh_hops+1, type, proto, table);
			}
		}

		return(0);
	}

	/* logging routing message */
	if(nlh->nlmsg_type == RTM_NEWROUTE)
		rec_log("[%-12s] route added: %s%stype=%s proto=%s table=%s",
			msg_type, msg, metric, type, proto, table);
	else if(nlh->nlmsg_type == RTM_DELROUTE)
		rec_log("[%-12s] route deleted: %s%stype=%s proto=%s table=%s",
			msg_type, msg, metric, type, proto, table);

	return(0);
}

/*
 * convert a route message type from number to string
 */ 
void parse_n2s_rtm_type(unsigned char ntype, char *stype)
{
	switch(ntype) {
		case RTN_UNSPEC:
			strcpy(stype, "none");
			break;
		case RTN_UNICAST:
			strcpy(stype, "unicast");
			break;
		case RTN_LOCAL:
			strcpy(stype, "local");
			break;
		case RTN_BROADCAST:
			strcpy(stype, "broadcast");
			break;
		case RTN_ANYCAST:
			strcpy(stype, "anycast");
			break;
		case RTN_MULTICAST:
			strcpy(stype, "multicast");
			break;
		case RTN_BLACKHOLE:
			strcpy(stype, "blackhole");
			break;
		case RTN_UNREACHABLE:
			strcpy(stype, "unreachable");
			break;
		case RTN_PROHIBIT:
			strcpy(stype, "prohibit");
			break;
		case RTN_THROW:
			strcpy(stype, "throw");
			break;
		case RTN_NAT:
			strcpy(stype, "nat");
			break;
		case RTN_XRESOLVE:
			strcpy(stype, "external");
			break;
		default:
			sprintf(stype, "%d", ntype);
	}
}

/*
 * convert a route message protocol from number to string
 */
void parse_n2s_rtm_protocol(int nproto, char *sproto)
{
	switch(nproto) {
		case RTPROT_UNSPEC:
			strcpy(sproto, "none");
			break;
		case RTPROT_REDIRECT:
			strcpy(sproto, "redirect");
			break;
		case RTPROT_KERNEL:
			strcpy(sproto, "kernel");
			break;
		case RTPROT_BOOT:
			strcpy(sproto, "boot");
			break;
		case RTPROT_STATIC:
			strcpy(sproto, "static");
			break;
		case RTPROT_GATED:
			strcpy(sproto, "gated");
			break;
		case RTPROT_RA:
			strcpy(sproto, "ra");
			break;
		case RTPROT_MRT:
			strcpy(sproto, "ra");
			break;
		case RTPROT_ZEBRA:
			strcpy(sproto, "zebra");
			break;
		case RTPROT_BIRD:
			strcpy(sproto, "bird");
			break;
		case RTPROT_DNROUTED:
			strcpy(sproto, "dnrouted");
			break;
		case RTPROT_XORP:
			strcpy(sproto, "xorp");
			break;
		case RTPROT_NTK:
			strcpy(sproto, "ntk");
			break;
		case RTPROT_DHCP:
			strcpy(sproto, "dhcp");
			break;
		default:
			sprintf(sproto, "%d", nproto);
	}
}

/*
 * convert a route table id from number to string
 */
void parse_n2s_rtm_table(unsigned ntable, char *stable)
{
	switch(ntable) {
		case RT_TABLE_UNSPEC:
			strcpy(stable, "none");
			break;
		case RT_TABLE_COMPAT:
			strcpy(stable, "compat");
			break;
		case RT_TABLE_DEFAULT:
			strcpy(stable, "default");
			break;
		case RT_TABLE_MAIN:
			strcpy(stable, "main");
			break;
		case RT_TABLE_LOCAL:
			strcpy(stable, "local");
			break;
		default:
			sprintf(stable, "%d", ntable);
	}
}

/*
 * debug route messages
 */
void debug_rtmsg(struct rtmsg *rtm, struct rtattr *rta[], int rtm_len)
{
	/* debug rtmsg */
	char rtm_flags_list[MAX_STR_SIZE] = "";
	rec_dbg("*********************************************************************");
	rec_dbg("[ rtmsg(%d) ]",
		NLMSG_ALIGN(sizeof(struct rtmsg)));
	rec_dbg("    rtm_family(%d): %d(%s)",
		sizeof(rtm->rtm_family), rtm->rtm_family,
		debug_n2s_af_type(rtm->rtm_family));
	rec_dbg("    rtm_dst_len(%d): %d",
		sizeof(rtm->rtm_dst_len), rtm->rtm_dst_len);
	rec_dbg("    rtm_src_len(%d): %d",
		sizeof(rtm->rtm_src_len), rtm->rtm_src_len);
	rec_dbg("    rtm_tos(%d): %d",
		sizeof(rtm->rtm_tos), rtm->rtm_tos);
	rec_dbg("    rtm_table(%d): %d(%s)",
		sizeof(rtm->rtm_table), rtm->rtm_table,
		debug_n2s_rtm_table(rtm->rtm_table));
	rec_dbg("    rtm_protocol(%d): %d(%s)",
		sizeof(rtm->rtm_protocol), rtm->rtm_protocol,
		debug_n2s_rtm_protocol(rtm->rtm_protocol));
	rec_dbg("    rtm_scope(%d): %d(%s)",
		sizeof(rtm->rtm_scope), rtm->rtm_scope,
		debug_n2s_rtm_scope(rtm->rtm_scope));
	rec_dbg("    rtm_type(%d): %d(%s)",
		sizeof(rtm->rtm_type), rtm->rtm_type,
		debug_n2s_rtm_type(rtm->rtm_type));
	debug_n2s_rtm_flags(rtm->rtm_flags, rtm_flags_list);
	rec_dbg("    rtm_flags(%d): %d(%s)",
		sizeof(rtm->rtm_flags), rtm->rtm_flags,
		rtm_flags_list);

	/* debug route attributes */
	rec_dbg("*********************************************************************");
	rec_dbg("[ rtmsg attributes(%d) ]",
		NLMSG_ALIGN(rtm_len - NLMSG_ALIGN(sizeof(struct rtmsg))));

	/* debug destination prefix */
	if(rta[RTA_DST]) {
		char dst[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_DST]), dst, sizeof(dst));
		rec_dbg("    RTA_DST(%hu): %s",
			RTA_ALIGN(rta[RTA_DST]->rta_len), dst);
	}

	/* debug source prefix */
	if(rta[RTA_SRC]) {
		char src[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_SRC]), src, sizeof(src));
		rec_dbg("    RTA_SRC(%hu): %s",
			RTA_ALIGN(rta[RTA_SRC]->rta_len), src);
	}

	/* debug ingress interface */
	if(rta[RTA_IIF]) {
		char iifname[IFNAMSIZ] = "";
		int iifindex = *((int *)RTA_DATA(rta[RTA_IIF]));
		if_indextoname(iifindex, iifname);
		rec_dbg("    RTA_IIF(%hu): %d(%s)",
			RTA_ALIGN(rta[RTA_IIF]->rta_len), iifindex, iifname);
	}

	/* debug egress interface */
	if(rta[RTA_OIF]) {
		char oifname[IFNAMSIZ] = "";
		int oifindex = *((int *)RTA_DATA(rta[RTA_OIF]));
		if_indextoname(oifindex, oifname);
		rec_dbg("    RTA_OIF(%hu): %d(%s)",
			RTA_ALIGN(rta[RTA_OIF]->rta_len), oifindex, oifname);
	}

	/* debug gateway */
	if(rta[RTA_GATEWAY]) {
		char gw[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_GATEWAY]), gw, sizeof(gw));
		rec_dbg("    RTA_GATEWAY(%hu): %s",
			RTA_ALIGN(rta[RTA_GATEWAY]->rta_len), gw);
	}

	/* debug priority(but metric) */
	if(rta[RTA_PRIORITY]) {
		int priority = *((int *)RTA_DATA(rta[RTA_PRIORITY]));
		rec_dbg("    RTA_PRIORITY(%hu): %d",
			RTA_ALIGN(rta[RTA_PRIORITY]->rta_len), priority);
	}

	/* debug preferred source address */
	if(rta[RTA_PREFSRC]) {
		char prefsrc[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_PREFSRC]), prefsrc, sizeof(prefsrc));
		rec_dbg("    RTA_PREFSRC(%hu): %s",
			RTA_ALIGN(rta[RTA_PREFSRC]->rta_len), prefsrc);
	}

	/* debug metric */
	if(rta[RTA_METRICS]) {
		struct rtattr *rtax[RTAX_MAX+1];
		rec_dbg("    RTA_METRICS(%hu):",
			RTA_ALIGN(rta[RTA_METRICS]->rta_len));
		parse_rtattr(rtax, RTAX_MAX, RTA_DATA(rta[RTA_METRICS]), RTA_PAYLOAD(rta[RTA_METRICS]));
		if(rtax[RTAX_LOCK])
			rec_dbg("        RTAX_LOCK(%hu): %u",
				RTA_ALIGN(rtax[RTAX_LOCK]->rta_len), *(unsigned *)RTA_DATA(rtax[RTAX_LOCK]));
		if(rtax[RTAX_MTU])
			rec_dbg("        RTAX_MTU(%hu): %u",
				RTA_ALIGN(rtax[RTAX_MTU]->rta_len), *(unsigned *)RTA_DATA(rtax[RTAX_MTU]));
		if(rtax[RTAX_ADVMSS])
			rec_dbg("        RTAX_ADVMSS(%hu): %u",
				RTA_ALIGN(rtax[RTAX_ADVMSS]->rta_len), *(unsigned *)RTA_DATA(rtax[RTAX_ADVMSS]));
		if(rtax[RTAX_HOPLIMIT])
			rec_dbg("        RTAX_HOPLIMIT(%hu): %d",
				RTA_ALIGN(rtax[RTAX_HOPLIMIT]->rta_len), *(int *)RTA_DATA(rtax[RTAX_HOPLIMIT]));
		if(rtax[RTAX_WINDOW])
			rec_dbg("        RTAX_WINDOW(%hu): %u",
				RTA_ALIGN(rtax[RTAX_WINDOW]->rta_len), *(unsigned *)RTA_DATA(rtax[RTAX_WINDOW]));
	}

	/* debug multipath */
	if(rta[RTA_MULTIPATH]) {
		struct rtnexthop *rtnh = RTA_DATA(rta[RTA_MULTIPATH]);
		int rtnh_len = RTA_PAYLOAD(rta[RTA_MULTIPATH]);
		struct rtattr *rtnha[__RTA_MAX];
		char rtnh_ifname[IFNAMSIZ] = "";
		char rtnh_gw[INET6_ADDRSTRLEN] = "";
		char rtnh_flags_list[MAX_STR_SIZE] = "";
		rec_dbg("    RTA_MULTIPATH(%hu):",
			RTA_ALIGN(rta[RTA_MULTIPATH]->rta_len));
		for(; RTNH_OK(rtnh, rtnh_len);
			rtnh = RTNH_NEXT(rtnh), rtnh_len -= RTNH_ALIGN(rtnh->rtnh_len)) {
			rec_dbg("        [ rtnexthop(%d) ]",
				NLMSG_ALIGN(sizeof(struct rtnexthop)));
			rec_dbg("            rtnh_len(%d): %hu",
				sizeof(rtnh->rtnh_len), rtnh->rtnh_len);
			debug_n2s_rtnh_flags(rtnh->rtnh_flags, rtnh_flags_list);
			rec_dbg("            rtnh_flags(%d): %d(%s)",
				sizeof(rtnh->rtnh_flags), rtnh->rtnh_flags,
				rtnh_flags_list);
			rec_dbg("            rtnh_hops(%d): %d",
				sizeof(rtnh->rtnh_hops), rtnh->rtnh_hops);
			if_indextoname(rtnh->rtnh_ifindex, rtnh_ifname);
			rec_dbg("            rtnh_ifindex(%d): %d(%s)",
				sizeof(rtnh->rtnh_ifindex), rtnh->rtnh_ifindex, rtnh_ifname);
			parse_rtattr(rtnha, RTA_MAX, RTNH_DATA(rtnh), rtnh->rtnh_len - sizeof(*rtnh));
			if(rtnha[RTA_GATEWAY]) {
				inet_ntop(rtm->rtm_family, RTA_DATA(rtnha[RTA_GATEWAY]),
					rtnh_gw, sizeof(rtnh_gw));
				rec_dbg("        RTA_GATEWAY(%hu): %s",
					RTA_ALIGN(rtnha[RTA_GATEWAY]->rta_len), rtnh_gw);
			}
		}
	}

	/* debug flow */
	if(rta[RTA_FLOW])
		rec_dbg("    RTA_FLOW(%hu): %u",
			RTA_ALIGN(rta[RTA_FLOW]->rta_len), *((unsigned *)RTA_DATA(rta[RTA_FLOW])));

	/* debug route cache information */
	if(rta[RTA_CACHEINFO]) {
		struct rta_cacheinfo *rtac;
		rtac = (struct rta_cacheinfo *)RTA_DATA(rta[RTA_CACHEINFO]);
		rec_dbg("    RTA_CACHEINFO(%hu):",
			RTA_ALIGN(rta[RTA_CACHEINFO]->rta_len));
		rec_dbg("        [ rta_cacheinfo(%d) ]",
			NLMSG_ALIGN(sizeof(struct rta_cacheinfo)));
		rec_dbg("            rta_clntref(%d): %u",
			sizeof(rtac->rta_clntref), rtac->rta_clntref);
		rec_dbg("            rta_lastuse(%d): %u",
			sizeof(rtac->rta_lastuse), rtac->rta_lastuse);
		rec_dbg("            rta_expires(%d): %d",
			sizeof(rtac->rta_expires), rtac->rta_expires);
		rec_dbg("            rta_error(%d): %u",
			sizeof(rtac->rta_error), rtac->rta_error);
		rec_dbg("            rta_used(%d): %u",
			sizeof(rtac->rta_used), rtac->rta_used);
		rec_dbg("            rta_id(%d): %u",
			sizeof(rtac->rta_id), rtac->rta_id);
		rec_dbg("            rta_ts(%d): %u",
			sizeof(rtac->rta_ts), rtac->rta_ts);
		rec_dbg("            rta_tsage(%d): %u",
			sizeof(rtac->rta_tsage), rtac->rta_tsage);
	}

	/* debug route table id & name */
	if(rta[RTA_TABLE])
		rec_dbg("    RTA_TABLE(%hu): %d(%s)",
			RTA_ALIGN(rta[RTA_TABLE]->rta_len), *(int *)RTA_DATA(rta[RTA_TABLE]),
			debug_n2s_rtm_table(*(int *)RTA_DATA(rta[RTA_TABLE])));

	rec_dbg("");

	return;
}

/*
 * convert rtm_table from number to string
 */
const char *debug_n2s_rtm_table(int table)
{
#define _RTM_TABLE(s) \
	if(table == RT_TABLE_##s) \
		return #s;
	_RTM_TABLE(UNSPEC);
	_RTM_TABLE(COMPAT);
	_RTM_TABLE(DEFAULT);
	_RTM_TABLE(MAIN);
	_RTM_TABLE(LOCAL);
#undef _RTM_TABLE
}

/*
 * convert rtm_protocol from number to string
 */
const char *debug_n2s_rtm_protocol(int protocol)
{
#define _RTM_PROTOCOL(s) \
	if(protocol == RTPROT_##s) \
		return #s;
	_RTM_PROTOCOL(UNSPEC);
	_RTM_PROTOCOL(REDIRECT);
	_RTM_PROTOCOL(KERNEL);
	_RTM_PROTOCOL(BOOT);
	_RTM_PROTOCOL(STATIC);
	_RTM_PROTOCOL(GATED);
	_RTM_PROTOCOL(RA);
	_RTM_PROTOCOL(MRT);
	_RTM_PROTOCOL(ZEBRA);
	_RTM_PROTOCOL(BIRD);
	_RTM_PROTOCOL(DNROUTED);
	_RTM_PROTOCOL(XORP);
	_RTM_PROTOCOL(NTK);
	_RTM_PROTOCOL(DHCP);
#undef _RTM_PROTOCOL
}

/*
 * convert rtm_scope from number to string
 */
const char *debug_n2s_rtm_scope(int scope)
{
#define _RTM_SCOPE(s) \
	if(scope == RT_SCOPE_##s) \
		return #s;
	_RTM_SCOPE(UNIVERSE);
	_RTM_SCOPE(SITE);
	_RTM_SCOPE(LINK);
	_RTM_SCOPE(HOST);
	_RTM_SCOPE(NOWHERE);
#undef _RTM_SCOPE
}

/*
 * convert rtm_type from number to string
 */
const char *debug_n2s_rtm_type(int type)
{
#define _RTM_TYPE(s) \
	if(type == RTN_##s) \
		return #s;
	_RTM_TYPE(UNSPEC);
	_RTM_TYPE(UNICAST);
	_RTM_TYPE(LOCAL);
	_RTM_TYPE(BROADCAST);
	_RTM_TYPE(ANYCAST);
	_RTM_TYPE(MULTICAST);
	_RTM_TYPE(BLACKHOLE);
	_RTM_TYPE(UNREACHABLE);
	_RTM_TYPE(PROHIBIT);
	_RTM_TYPE(THROW);
	_RTM_TYPE(NAT);
	_RTM_TYPE(XRESOLVE);
#undef _RTM_TYPE
}

/*
 * convert rtm_flags from number to string
 */
void debug_n2s_rtm_flags(int flags, char *flags_list)
{
#define _RTM_FLAGS(s) \
	if(flags & RTM_F_##s) \
		(flags &= ~RTM_F_##s) ? strcat(flags_list, #s ", ") : strcat(flags_list, #s);
	_RTM_FLAGS(NOTIFY);
	_RTM_FLAGS(CLONED);
	_RTM_FLAGS(EQUALIZE);
	_RTM_FLAGS(PREFIX);
#undef _RTM_FLAGS
}

/*
 * convert rtnh_flags from number to string
 */
void debug_n2s_rtnh_flags(int flags, char *flags_list)
{
#define _RTNH_FLAGS(s) \
	if(flags & RTNH_F_##s) \
		(flags &= ~RTNH_F_##s) ? strcat(flags_list, #s ", ") : strcat(flags_list, #s);
	_RTNH_FLAGS(DEAD);
	_RTNH_FLAGS(PERVASIVE);
	_RTNH_FLAGS(ONLINK);
#undef _RTNH_FLAGS
}
